#!/usr/bin/env python3

from ..error_logging import fail_guess_because
from ..string_manglers import mangle_suffixes_or_die


def stub_all_new_para(wordmap, stubmap):
    if not wordmap['new_para']:
        return wordmap
    if not wordmap['new_para'] in stubmap:
        fail_guess_because(wordmap, "", wordmap['new_para'],
                           "Missing stub paradigm" +
                           ": Update stub-deletions.tsv and carry on")
        return None
    elif not stubmap[wordmap['new_para']] or \
            stubmap[wordmap['new_para']] == '':
        return wordmap
    wordmap = mangle_suffixes_or_die(wordmap, [stubmap[wordmap['new_para']]])
    return wordmap


def stub_all_ktn(wordmap):
    '''Generate unmodifiable stub for inflectional processes.
    this cuts every morphologically varying character.
    '''
    if not wordmap['kotus_tn']:
        return wordmap
    if wordmap['pos'] == 'ACRONYM':
        return wordmap
    tn = int(wordmap['kotus_tn'])
    if not wordmap['kotus_av'] or tn in [28, 60, 1007, 1009, 1010]:
        if tn in [1, 2, 3, 4, 8, 21, 22, 48]:
            pass
        elif tn in [5, 6]:
            if wordmap['stub'].endswith('i'):
                wordmap = mangle_suffixes_or_die(wordmap, ['i'])
            else:
                pass
        elif tn == 10:
            if wordmap['stub'].endswith('n'):
                wordmap = mangle_suffixes_or_die(wordmap, ['an', 'än'])
            elif not wordmap['stub'].endswith('A'):
                wordmap = mangle_suffixes_or_die(wordmap, ['a', 'ä', 'ă'])
        elif tn == 41 and wordmap['pos'] == 'ADJECTIVE'  \
                and (wordmap['stub'].endswith('is') or
                     wordmap['stub'].endswith('paras')):
            wordmap = mangle_suffixes_or_die(wordmap, ['is', 'as'])
        elif tn == 7:
            wordmap = mangle_suffixes_or_die(wordmap, ['i'])
        elif tn in range(9, 16):
            if not wordmap['stub'].endswith('A'):
                wordmap = mangle_suffixes_or_die(wordmap, ['a', 'ä', 'e'])
        elif tn in [16, 23, 24, 26]:
            wordmap = mangle_suffixes_or_die(wordmap, ['i'])
        elif tn in [17, 18, 20]:
            wordmap = mangle_suffixes_or_die(wordmap, ['a', 'e', 'i', 'o',
                                                       'u', 'y', 'ä', 'ö'])
        elif tn == 19:
            wordmap = mangle_suffixes_or_die(wordmap, ['uo', 'yö', 'ie'])
        elif tn == 25:
            wordmap = mangle_suffixes_or_die(wordmap, ['mi'])
        elif tn in [27, 28]:
            wordmap = mangle_suffixes_or_die(wordmap, ['si', 's'])
        elif tn == 29:
            wordmap = mangle_suffixes_or_die(wordmap, ['psi', 'ksi'])
        elif tn == 30:
            wordmap = mangle_suffixes_or_die(wordmap, ['tsi'])
        elif tn == 31:
            wordmap = mangle_suffixes_or_die(wordmap, ['ksi'])
        elif tn == 32:
            if wordmap['stub'].endswith('kymmenen'):
                wordmap = mangle_suffixes_or_die(wordmap, ['en'])
            else:
                pass
        elif tn in range(33, 38):
            wordmap = mangle_suffixes_or_die(wordmap, ['n'])
        elif tn == 38:
            wordmap = mangle_suffixes_or_die(wordmap, ['nen'])
        elif tn in [39, 41, 42]:
            wordmap = mangle_suffixes_or_die(wordmap, ['s'])
        elif tn == 40:
            wordmap = mangle_suffixes_or_die(wordmap, ['s', 't'])
        elif tn in [43, 44, 46]:
            wordmap = mangle_suffixes_or_die(wordmap, ['t'])
        elif tn == 45:
            wordmap = mangle_suffixes_or_die(wordmap, ['s'])
        elif tn == 47:
            wordmap = mangle_suffixes_or_die(wordmap, ['ut', 'yt'])
        elif tn == 49:
            if wordmap['stub'].endswith('e'):
                wordmap['extra_e'] = True
        elif tn == 52:
            wordmap['stub'] = wordmap['stub'][:-1]
        elif tn in [53, 56, 77]:
            wordmap['stub'] = wordmap['stub'][:-2]
        elif tn == 54:
            wordmap['stub'] = wordmap['stub'][:-3]
        elif tn in [55, 57, 76]:
            wordmap['stub'] = wordmap['stub'][:-3]
        elif tn == 58:
            wordmap['stub'] = wordmap['stub'][:-2]
        elif tn == 59:
            wordmap['stub'] = wordmap['stub'][:-3]
        elif tn == 60:
            wordmap['stub'] = wordmap['stub'][:-4]
        elif tn == 61:
            wordmap['stub'] = wordmap['stub'][:-2]
        elif tn in [62, 68]:
            wordmap['stub'] = wordmap['stub'][:-3]
        elif tn in [62, 63, 65]:
            wordmap['stub'] = wordmap['stub'][:-3]
        elif tn == 64:
            wordmap['stub'] = wordmap['stub'][:-4]
        elif tn == 67 and wordmap['stub'] in ['mennä', 'panna', 'tulla']:
            wordmap['stub'] = wordmap['stub'][:-3]
        elif tn == 67 and wordmap['stub'][-5:] in ['ailla', 'äillä', 'oilla',
                                                   'öillä']:
            wordmap['stub'] = wordmap['stub'][:-4]
        elif tn in [66, 67, 69, 72, 73, 74, 75]:
            wordmap['stub'] = wordmap['stub'][:-2]
        elif tn in [70, 71]:
            wordmap['stub'] = wordmap['stub'][:-3]
        elif tn == 78:
            wordmap = mangle_suffixes_or_die(wordmap, ['ta', 'tä', 'a', 'ä'])
        elif tn == 1007:
            wordmap = mangle_suffixes_or_die(wordmap, ['i'])
        elif tn == 1008:
            wordmap = mangle_suffixes_or_die(wordmap, ['e'])
        elif tn in [1010, 1009]:
            wordmap = mangle_suffixes_or_die(wordmap, ['ika', 'tkä'])
        elif tn in [1024, 1026]:
            wordmap = mangle_suffixes_or_die(wordmap, ['i'])
        elif tn == 1067:
            wordmap = mangle_suffixes_or_die(wordmap, ['olla'])
        elif tn == 1099:
            wordmap = mangle_suffixes_or_die(wordmap, ['ei'])
        elif wordmap['kotus_tn'] == 101:
            if wordmap['lemma'] in ['minä', 'sinä']:
                wordmap = mangle_suffixes_or_die(wordmap, ['nä'])
            elif wordmap['lemma'] == 'hän':
                pass
            elif wordmap['lemma'] in ['me', 'te', 'he']:
                pass
            elif wordmap['lemma'] == 'tämä':
                wordmap = mangle_suffixes_or_die(wordmap, ['mä'])
            elif wordmap['lemma'] == 'tuo':
                wordmap = mangle_suffixes_or_die(wordmap, ['uo'])
            elif wordmap['lemma'] == 'se':
                wordmap = mangle_suffixes_or_die(wordmap, ['e'])
            elif wordmap['lemma'] == 'nämä':
                wordmap = mangle_suffixes_or_die(wordmap, ['mä'])
            elif wordmap['lemma'] == 'nuo':
                wordmap = mangle_suffixes_or_die(wordmap, ['uo'])
            elif wordmap['lemma'] == 'ne':
                wordmap = mangle_suffixes_or_die(wordmap, ['e'])
            elif wordmap['lemma'] == 'joku':
                wordmap = mangle_suffixes_or_die(wordmap, ['ku'])
            elif wordmap['lemma'] == 'joka':
                wordmap = mangle_suffixes_or_die(wordmap, ['ka'])
            elif wordmap['lemma'] == 'jokin':
                wordmap = mangle_suffixes_or_die(wordmap, ['kin'])
            elif wordmap['lemma'] == 'jompikumpi':
                wordmap = mangle_suffixes_or_die(wordmap, ['pikumpi'])
            elif wordmap['lemma'] == 'kuka':
                wordmap = mangle_suffixes_or_die(wordmap, ['uka'])
            elif wordmap['lemma'] == 'kukaan':
                wordmap = mangle_suffixes_or_die(wordmap, ['ukaan'])
            elif wordmap['lemma'] == 'kukakin':
                wordmap = mangle_suffixes_or_die(wordmap, ['ukakin'])
            elif wordmap['lemma'].endswith('kukin'):
                wordmap = mangle_suffixes_or_die(wordmap, ['ukin'])
            elif wordmap['lemma'] == 'mikin':
                wordmap = mangle_suffixes_or_die(wordmap, ['kin'])
            elif wordmap['lemma'] == 'mikä':
                wordmap = mangle_suffixes_or_die(wordmap, ['kä'])
            elif wordmap['lemma'] == 'mikäkin':
                wordmap = mangle_suffixes_or_die(wordmap, ['käkin'])
            elif wordmap['lemma'] == 'mikään':
                wordmap = mangle_suffixes_or_die(wordmap, ['kään'])
            elif wordmap['lemma'] == 'missä':
                wordmap = mangle_suffixes_or_die(wordmap, ['ssä'])
            elif wordmap['lemma'] == 'missäkään':
                wordmap = mangle_suffixes_or_die(wordmap, ['ssäkään'])
            elif wordmap['lemma'] == 'missään':
                wordmap = mangle_suffixes_or_die(wordmap, ['ssään'])
            elif wordmap['lemma'] in ['muuan', 'muudan']:
                wordmap = mangle_suffixes_or_die(wordmap, ['n'])
            elif wordmap['lemma'] in ['mä', 'sä']:
                wordmap = mangle_suffixes_or_die(wordmap, ['ä'])
            elif wordmap['lemma'] in ['mie', 'sie']:
                wordmap = mangle_suffixes_or_die(wordmap, ['e'])
            elif wordmap['lemma'] == 'toi':
                wordmap = mangle_suffixes_or_die(wordmap, ['i'])
            elif wordmap['lemma'].endswith('ainoa'):
                wordmap = mangle_suffixes_or_die(wordmap, ['a'])
            elif wordmap['lemma'] == 'kumpikaan':
                wordmap = mangle_suffixes_or_die(wordmap, ['pikaan'])
            elif wordmap['lemma'].endswith('kumpikin'):
                wordmap = mangle_suffixes_or_die(wordmap, ['pikin'])
            elif wordmap['lemma'] in ['jota', 'kenkään', 'kuta', 'ma', 'mi',
                                      'missäkin', 'monta', 'montaa', 'sa',
                                      'tää', 'ken', 'koko']:
                pass
        elif wordmap['kotus_tn'] in [99, 999] and \
                wordmap['possessive'] == 'optional':
            if wordmap['stub'].endswith('den'):
                wordmap = mangle_suffixes_or_die(
                    wordmap, ['den'])  # näh|te| -mme
            elif wordmap['stub'].endswith('n'):
                wordmap = mangle_suffixes_or_die(
                    wordmap, ['n'])  # näkyvii|n -mme
            elif wordmap['stub'].endswith('i'):
                wordmap = mangle_suffixes_or_die(
                    wordmap, ['i'])  # vuoks|i -e-mme
        elif wordmap['kotus_tn'] in [99, 999] and \
                wordmap['possessive'] == 'obligatory':
            if wordmap['stub'].endswith('n'):
                wordmap['stub'] = wordmap['stub'][:-2]   # hyvillä|än -mme
            elif wordmap['stub'].endswith('nsa') or \
                    wordmap['stub'].endswith('nsä'):
                wordmap['stub'] = wordmap['stub'][:-3]   # aika|nsa -mme
        elif wordmap['kotus_tn'] in [0, 99, 999]:
            pass
        elif wordmap['kotus_tn'] == 1101:
            wordmap = mangle_suffixes_or_die(wordmap, ['ka', 'kä'])
        elif wordmap['kotus_tn'] == 51:
            wordmap['stub'] = ''
        else:
            fail_guess_because(wordmap, ['!av'], ['0-71', 999, 1007, 1010,
                                                  1009, 1024, 1026, 1067,
                                                  1099])
    elif wordmap['grade_dir'] == 'weaken':
        if wordmap['kotus_av'] == 'D' and wordmap['lemma'].endswith('uoka'):
            wordmap = mangle_suffixes_or_die(wordmap, ['oka'])
        elif wordmap['kotus_av'] in ['A', 'D', 'G', 'L', 'M']:
            if tn in [1, 2, 3, 4]:
                wordmap = mangle_suffixes_or_die(wordmap, ['ko', 'ku',
                                                           'ky', 'kö'])
            elif tn in [5, 6, 7]:
                wordmap = mangle_suffixes_or_die(wordmap, ['ki'])
            elif tn in [8]:
                wordmap = mangle_suffixes_or_die(wordmap, ['ke'])
            elif tn in [9, 10, 11, 12, 13, 14]:
                wordmap = mangle_suffixes_or_die(wordmap, ['ka', 'kä'])
            elif tn in range(52, 79):
                lastk = wordmap['stub'].rfind('k')
                wordmap['stub'] = wordmap['stub'][:lastk]
            else:
                fail_guess_because(wordmap, ['av-', 'A', 'D', 'G', 'L', 'M'],
                                   ['1-14'])
        elif wordmap['kotus_av'] in ['B', 'E', 'H']:
            if tn in [1, 2, 3, 4]:
                wordmap = mangle_suffixes_or_die(wordmap, ['po', 'pu',
                                                           'py', 'pö'])
            elif tn in [5, 6, 7, 16]:
                wordmap = mangle_suffixes_or_die(wordmap, ['pi'])
            elif tn in [8]:
                wordmap = mangle_suffixes_or_die(wordmap, ['pe'])
            elif tn in [9, 10, 11, 12, 13, 14]:
                wordmap = mangle_suffixes_or_die(wordmap, ['pa', 'pä'])
            elif tn in range(52, 79):
                lastp = wordmap['stub'].rfind('p')
                wordmap['stub'] = wordmap['stub'][:lastp]
            else:
                fail_guess_because(wordmap, ['av-', 'B', 'E', 'H'],
                                   ['1-14'])
        elif wordmap['kotus_av'] in ['C', 'F', 'I', 'J', 'K']:
            if tn in [1, 2, 3, 4]:
                wordmap = mangle_suffixes_or_die(wordmap, ['to', 'tu',
                                                           'ty', 'tö'])
            elif tn in [5, 6, 7]:
                wordmap = mangle_suffixes_or_die(wordmap, ['ti'])
            elif tn in [8]:
                wordmap = mangle_suffixes_or_die(wordmap, ['te'])
            elif tn in [9, 10, 11, 12, 13, 14]:
                wordmap = mangle_suffixes_or_die(wordmap, ['ta', 'tä'])
            elif tn in range(52, 79):
                lastt = wordmap['stub'].rfind('t')
                wordmap['stub'] = wordmap['stub'][:lastt]
            else:
                fail_guess_because(wordmap, ['av-', 'C', 'F', 'I', 'J', 'K'],
                                   ['1-14'])
        else:
            fail_guess_because(wordmap, ['av+'], ['A-T'])
    elif wordmap['grade_dir'] == 'strengthen':
        if wordmap['kotus_av'] == 'A':
            if tn in [33]:
                wordmap = mangle_suffixes_or_die(wordmap, ['in'])
            elif tn in [41]:
                wordmap = mangle_suffixes_or_die(wordmap, ['as', 'äs'])
            elif tn in [48]:
                wordmap = mangle_suffixes_or_die(wordmap, ['e'])
            elif tn in range(52, 79):
                lastk = wordmap['stub'].rfind('k')
                wordmap['stub'] = wordmap['stub'][:lastk + 1]
            else:
                fail_guess_because(wordmap, ['av+', 'A'], [33, 41, 48])
        elif wordmap['kotus_av'] == 'B':
            if tn in [33, 34]:
                wordmap = mangle_suffixes_or_die(wordmap, ['an'])
            elif tn in [41]:
                wordmap = mangle_suffixes_or_die(wordmap, ['as', 'äs', 'es'])
            elif tn in [48]:
                wordmap = mangle_suffixes_or_die(wordmap, ['e'])
            elif tn in range(52, 79):
                lastp = wordmap['stub'].rfind('p')
                wordmap['stub'] = wordmap['stub'][:lastp + 1]
            else:
                fail_guess_because(wordmap, ['av+', 'B'], [33, 41, 48])
        elif wordmap['kotus_av'] == 'C':
            if tn == 32:
                wordmap = mangle_suffixes_or_die(wordmap, ['ar', 'är'])
            elif tn == 33:
                wordmap = mangle_suffixes_or_die(wordmap, ['in'])
            elif tn == 34:
                wordmap = mangle_suffixes_or_die(
                    wordmap, ['on', 'ön', 'oin', 'öin'])
            elif tn == 41:
                wordmap = mangle_suffixes_or_die(wordmap, ['as', 'us', 'is',
                                                           'es', 'äs'])
            elif tn == 44:
                wordmap = mangle_suffixes_or_die(wordmap, ['et'])
            elif tn == 48:
                wordmap = mangle_suffixes_or_die(wordmap, ['e'])
            elif tn in range(52, 79):
                lastt = wordmap['stub'].rfind('t')
                if wordmap['kotus_tn'] == 73:
                    lastt = wordmap['stub'].rfind('t', 0, lastt)
                wordmap['stub'] = wordmap['stub'][:lastt + 1]
            else:
                fail_guess_because(wordmap, ['av+', 'C'], [32, 33, 34,
                                                           41, 44, 48])
        elif wordmap['kotus_av'] == 'D':
            # DANGER TERROR HORROR !!!!!!
            if wordmap['kotus_tn'] == 32:
                wordmap = mangle_suffixes_or_die(wordmap, ['en'])
            elif wordmap['kotus_tn'] == 33:
                wordmap = mangle_suffixes_or_die(wordmap, ['in'])
            elif wordmap['kotus_tn'] == 41:
                wordmap = mangle_suffixes_or_die(wordmap,
                                                 ['as', 'es', 'is', 'äs'])
            elif wordmap['kotus_tn'] == 43:
                wordmap = mangle_suffixes_or_die(wordmap, ['ut', 'yt'])
            elif wordmap['kotus_tn'] == 44:
                wordmap = mangle_suffixes_or_die(wordmap, ['et'])
            elif wordmap['kotus_tn'] == 48:
                wordmap = mangle_suffixes_or_die(wordmap, ['e'])
            elif wordmap['kotus_tn'] == 49:
                wordmap = mangle_suffixes_or_die(wordmap, ['en'])
            elif wordmap['kotus_tn'] == 67:
                wordmap = mangle_suffixes_or_die(wordmap,
                                                 ['ella', 'illa'])
            elif wordmap['kotus_tn'] == 72:
                wordmap = mangle_suffixes_or_die(wordmap,
                                                 ['ata', 'eta', 'etä', 'ota'])
            elif wordmap['kotus_tn'] == 73:
                wordmap = mangle_suffixes_or_die(wordmap,
                                                 ['ata', 'ätä'])
            elif wordmap['kotus_tn'] == 74:
                wordmap = mangle_suffixes_or_die(wordmap,
                                                 ['ota', 'eta', 'ötä', 'etä'])
            elif wordmap['kotus_tn'] == 75:
                wordmap = mangle_suffixes_or_die(wordmap,
                                                 ['ita', 'itä'])
            elif wordmap['kotus_tn'] == 99:
                pass
            else:
                fail_guess_because(wordmap, ['N', 'V', 'D'], [32, 33, 44, 48,
                                                              49, 67, 72, 73,
                                                              74, 75],
                                   "missing inverse 0:k gradation")
        elif wordmap['kotus_av'] == 'E':
            if tn == 33:
                wordmap = mangle_suffixes_or_die(wordmap, ['vin'])
            elif tn == 41:
                wordmap = mangle_suffixes_or_die(wordmap, ['vas', 'väs'])
            elif tn == 48:
                wordmap = mangle_suffixes_or_die(wordmap, ['ve'])
            elif wordmap['kotus_tn'] == 49:
                wordmap = mangle_suffixes_or_die(wordmap, ['val'])
            elif tn in range(52, 79):
                lastp = wordmap['stub'].rfind('v')
                wordmap['stub'] = wordmap['stub'][:lastp]
            else:
                fail_guess_because(wordmap, ['av+', 'E'], [33, 41, 48, 49])
        elif wordmap['kotus_av'] == 'F':
            if tn == 33:
                wordmap = mangle_suffixes_or_die(
                    wordmap, ['don', 'dun', 'din'])
            elif tn == 41:
                wordmap = mangle_suffixes_or_die(
                    wordmap, ['das', 'däs', 'des'])
            elif tn == 44:
                wordmap = mangle_suffixes_or_die(wordmap, ['det'])
            elif tn == 48:
                wordmap = mangle_suffixes_or_die(wordmap, ['de'])
            elif wordmap['kotus_tn'] == 49:
                wordmap = mangle_suffixes_or_die(wordmap, ['dar'])
            elif tn in range(52, 79):
                lastt = wordmap['stub'].rfind('d')
                wordmap['stub'] = wordmap['stub'][:lastt]
            else:
                fail_guess_because(wordmap, ['av+', 'F'], [33, 41, 44, 48, 49])
        elif wordmap['kotus_av'] == 'G':
            if tn == 41:
                wordmap = mangle_suffixes_or_die(wordmap, ['gas', 'gäs'])
            elif tn == 44:
                wordmap = mangle_suffixes_or_die(wordmap, ['get'])
            elif tn == 48:
                wordmap = mangle_suffixes_or_die(wordmap, ['ge'])
            elif wordmap['kotus_tn'] == 49:
                wordmap = mangle_suffixes_or_die(wordmap, ['ger'])
            elif tn in range(52, 79):
                lastk = wordmap['stub'].rfind('g')
                wordmap['stub'] = wordmap['stub'][:lastk]
            else:
                fail_guess_because(wordmap, ['av+', 'G'], [41, 44, 48])
        elif wordmap['kotus_av'] == 'H':
            if tn == 35:
                wordmap = mangle_suffixes_or_die(wordmap, ['min'])
            elif tn == 41:
                wordmap = mangle_suffixes_or_die(wordmap, ['mas', 'mäs'])
            elif tn == 43:
                wordmap = mangle_suffixes_or_die(wordmap, ['myt'])
            elif tn == 48:
                wordmap = mangle_suffixes_or_die(wordmap, ['me'])
            elif wordmap['kotus_tn'] == 49:
                wordmap = mangle_suffixes_or_die(wordmap, ['mel', 'mol'])
            elif tn in range(52, 79):
                lastp = wordmap['stub'].rfind('m')
                wordmap['stub'] = wordmap['stub'][:lastp]
            else:
                fail_guess_because(wordmap, ['av+', 'H'], [35, 41, 43, 48, 49])
        elif wordmap['kotus_av'] == 'I':
            if tn == 33:
                wordmap = mangle_suffixes_or_die(wordmap, ['lin'])
            elif tn == 41:
                wordmap = mangle_suffixes_or_die(wordmap, ['las', 'läs'])
            elif tn == 44:
                wordmap = mangle_suffixes_or_die(wordmap, ['let'])
            elif tn == 48:
                wordmap = mangle_suffixes_or_die(wordmap, ['le'])
            elif tn in range(52, 79):
                lastp = -1
                if wordmap['kotus_tn'] == 67:
                    lastp = wordmap['stub'].rfind('lell')
                else:
                    lastp = wordmap['stub'].rfind('l')
                wordmap['stub'] = wordmap['stub'][:lastp]
            else:
                fail_guess_because(wordmap, ['av+', 'I'], [33, 41, 44, 48])
        elif wordmap['kotus_av'] == 'J':
            if tn == 33:
                wordmap = mangle_suffixes_or_die(wordmap, ['nin'])
            elif tn == 41:
                wordmap = mangle_suffixes_or_die(wordmap, ['nas', 'näs'])
            elif tn == 44:
                wordmap = mangle_suffixes_or_die(wordmap, ['net'])
            elif tn == 48:
                wordmap = mangle_suffixes_or_die(wordmap, ['ne'])
            elif wordmap['kotus_tn'] == 49:
                wordmap = mangle_suffixes_or_die(
                    wordmap, ['nar', 'ner', 'nel'])
            elif tn in range(52, 79):
                lastp = wordmap['stub'].rfind('n')
                wordmap['stub'] = wordmap['stub'][:lastp]
            else:
                fail_guess_because(wordmap, ['av+', 'J'], [33, 41, 44, 48])
        elif wordmap['kotus_av'] == 'K':
            if tn == 33:
                wordmap = mangle_suffixes_or_die(wordmap, ['rin', 'roin'])
            elif tn == 41:
                wordmap = mangle_suffixes_or_die(wordmap, ['ras', 'räs'])
            elif tn == 44:
                wordmap = mangle_suffixes_or_die(wordmap, ['ret'])
            elif tn == 48:
                wordmap = mangle_suffixes_or_die(wordmap, ['re'])
            elif tn in range(52, 79):
                lastp = wordmap['stub'].rfind('r')
                wordmap['stub'] = wordmap['stub'][:lastp]
            else:
                fail_guess_because(wordmap, ['av+', 'K'], [33, 41, 44, 48])
        elif wordmap['kotus_av'] == 'L':
            if tn == 33:
                wordmap = mangle_suffixes_or_die(wordmap, ['jin'])
            elif tn == 48:
                wordmap = mangle_suffixes_or_die(wordmap, ['je'])
            elif tn in range(52, 79):
                lastp = wordmap['stub'].rfind('j')
                wordmap['stub'] = wordmap['stub'][:lastp]
            else:
                fail_guess_because(wordmap, ['av+', 'K'], [33, 48])
        elif wordmap['kotus_av'] == 'N':
            lastp = wordmap['stub'].rfind('d')
            wordmap['stub'] = wordmap['stub'][:lastp + 1]
            if lastp == -1:
                fail_guess_because(wordmap, ['av+', 'N'], ['d'])
        elif wordmap['kotus_av'] == 'P':
            lastp = wordmap['stub'].rfind('b')
            wordmap['stub'] = wordmap['stub'][:lastp + 1]
            if lastp == -1:
                fail_guess_because(wordmap, ['av+', 'P'], ['b'])
        elif wordmap['kotus_av'] == 'O':
            lastp = wordmap['stub'].rfind('g')
            wordmap['stub'] = wordmap['stub'][:lastp + 1]
            if lastp == -1:
                fail_guess_because(wordmap, ['av+', 'O'], ['g'])
        elif wordmap['kotus_av'] == 'T':
            lastp = wordmap['stub'].rfind('e')
            wordmap['stub'] = wordmap['stub'][:lastp]
            if lastp == -1:
                fail_guess_because(wordmap, ['av+', 'T'], ['auer'])
        else:
            fail_guess_because(wordmap, ['av-'], ['A-T'])
        return wordmap
    return wordmap


def stub_legacy(wordmap):
    '''Generate unmodifiable stub for inflectional processes.
    this cuts every morphologically varying character except gradating stop
    from ends of the words.
    '''
    tn = int(wordmap['kotus_tn'])
    # then cut
    if tn in range(1, 5) or tn in [8, 21, 22, 32, 48]:
        wordmap['stub'] = wordmap['stub']
    if tn in [5, 6]:
        if wordmap['extra_i']:
            wordmap['stub'] = wordmap['stub'][:-1]
        else:
            wordmap['stub'] = wordmap['stub']
    elif tn in [7, 16] or tn in range(33, 38) or tn in range(39, 47):
        wordmap['stub'] = wordmap['stub'][:-1]
    elif tn in range(9, 16):
        wordmap['stub'] = wordmap['stub'][:-1]
    elif tn in [16, 23, 24, 26]:
        wordmap['stub'] = wordmap['stub'].rstrip('i')
    elif tn in range(17, 19) or tn == 20:
        wordmap['stub'] = wordmap['stub'][:-1]
    elif tn == 19:
        wordmap['stub'] = wordmap['stub'][:-2]
    elif tn in [25, 27]:
        wordmap['stub'] = wordmap['stub'][:-2]
    elif tn == 28:
        wordmap['stub'] = wordmap['stub'][:-2]
    elif tn in [29, 30, 31, 38]:
        wordmap['stub'] = wordmap['stub'][:-3]
    elif tn == 47:
        wordmap['stub'] = wordmap['stub'][:-2]
    elif tn == 49:
        if wordmap['stub'].endswith('e'):
            wordmap['extra_e'] = True
    elif tn in [52, 78]:
        wordmap['stub'] = wordmap['stub'][:-1]
    elif tn in [53, 56, 77]:
        wordmap['stub'] = wordmap['stub'][:-2]
    elif tn == 54:
        if wordmap['kotus_av']:
            wordmap['stub'] = wordmap['stub'][:-7]
        else:
            wordmap['stub'] = wordmap['stub'][:-3]
    elif tn in [55, 57, 76]:
        if wordmap['kotus_av']:
            wordmap['stub'] = wordmap['stub'][:-7]  # 3 + gradation mark
        else:
            wordmap['stub'] = wordmap['stub'][:-3]
    elif tn == 58:
        wordmap['stub'] = wordmap['stub'][:-2]
    elif tn == 59:
        if wordmap['kotus_av']:
            wordmap['stub'] = wordmap['stub'][:-7]  # 3 + gradation mark
        else:
            wordmap['stub'] = wordmap['stub'][:-3]
    elif tn == 60:
        if wordmap['kotus_av']:
            wordmap['stub'] = wordmap['stub'][:-8]  # 4 + gradation mark
        else:
            wordmap['stub'] = wordmap['stub'][:-4]
    elif tn == 61:
        wordmap['stub'] = wordmap['stub'][:-2]
    elif tn in [62, 68]:
        wordmap['stub'] = wordmap['stub'][:-3]
    elif tn in [62, 63, 65]:
        wordmap['stub'] = wordmap['stub'][:-3]
    elif tn == 64:
        wordmap['stub'] = wordmap['stub'][:-4]
    elif tn in [66, 67, 1067, 69, 72, 73, 74, 75, 77]:
        wordmap['stub'] = wordmap['stub'][:-2]
    elif tn in [70, 71, 71]:
        wordmap['stub'] = wordmap['stub'][:-3]
    elif tn == 1007:
        wordmap['stub'] = wordmap['stub'][:-1]
    elif tn in [1010, 1009]:
        if wordmap['kotus_av']:
            wordmap['stub'] = wordmap['stub'][:-7]  # 3 + gradation mark
        else:
            wordmap['stub'] = wordmap['stub'][:-3]
    elif tn in [1024, 1026]:
        wordmap['stub'] = wordmap['stub'][:-1]
    return wordmap



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from argparse import ArgumentParser, FileType
# CLI stuff
from sys import stdin, stdout, stderr
# statistics
from time import perf_counter, process_time

import json

# omorfi
from omorfi import Omorfi


def main():
    """Invoke a simple CLI analyser."""
    a = ArgumentParser()
    a.add_argument('-a', '--analyser', metavar='AFILE',
                   help="load tokeniser model from (analyser) AFILE",
                   required=True)
    a.add_argument('-i', '--input', metavar="INFILE", type=open,
                   dest="infile", help="source of analysis data")
    a.add_argument('-v', '--verbose', action='store_true',
                   help="print verbosely while processing")
    a.add_argument('-o', '--output', metavar="OUTFILE", dest="outfile",
                   help="print output into OUTFILE", type=FileType('w'))
    a.add_argument('-x', '--statistics', metavar="STATFILE", dest="statfile",
                   help="print statistics to STATFILE", type=FileType('w'))
    a.add_argument('-O', '--output-format', metavar="OUTFORMAT",
                   default="moses",
                   help="format output for OUTFORMAT", choices=['moses',
                                                                'conllu',
                                                                'json',
                                                                'ftb3'])
    options = a.parse_args()
    omorfi = Omorfi(options.verbose)
    if options.analyser:
        if options.verbose:
            print("reading language model", options.analyser)
        omorfi.load_analyser(options.analyser)
    else:
        print("analyser is needed for tokenisation", file=stderr)
        exit(1)
    if not options.infile:
        options.infile = stdin
    if options.verbose:
        print("analysing", options.infile.name)
    if not options.outfile:
        options.outfile = stdout
    if options.verbose:
        print("writing to", options.outfile.name)
    if not options.statfile:
        options.statfile = stderr
    # statistics
    realstart = perf_counter()
    cpustart = process_time()
    tokens = 0
    lines = 0
    if options.output_format == 'conllu':
        print("# new doc id=", options.infile.name, file=options.outfile)
    for line in options.infile:
        line = line
        lines += 1
        if options.verbose and lines % 10000 == 0:
            print(lines, "...")
        if not line or line.rstrip('\n') == '':
            continue
        surfs = omorfi.tokenise(line)
        tokens += len(surfs)
        if options.output_format == 'moses':
            print(' '.join([surf.surf for surf in surfs]),
                  file=options.outfile)
        elif options.output_format == 'json':
            print(json.encode(surfs))
        elif options.output_format == 'conllu':
            print("# sent_id =", lines, file=options.outfile)
            print("# text =", line.rstrip("\n"), file=options.outfile)
            i = 1
            for surf in surfs:
                print(i, surf['surf'], "_", "_", "_", "_", "_", "_", "_", "_",
                      sep="\t", file=options.outfile)
                i += 1
        elif options.output_format == 'ftb3':
            print("<s><loc file=\"", options.infile.name, "\" line=\"",
                  lines, "\" />", file=options.outfile, sep="")
            i = 1
            for surf in surfs:
                print(i, surf['surf'], "_", "_", "_", "_", "_", "_", "_", "_",
                      sep="\t", file=options.outfile)
                i += 1
            print("</s>", file=options.outfile)
        if options.output_format == 'conllu':
            print(file=options.outfile)
    cpuend = process_time()
    realend = perf_counter()
    print("Lines:", lines, "Tokens:", tokens, "Ratio:", tokens / lines,
          "tokens/line", file=options.statfile)
    print("CPU time:", cpuend - cpustart, "Real time:", realend - realstart,
          file=options.statfile)
    print("Tokens per timeunit:", tokens / (realend - realstart),
          "Lines per timeunit:", lines / (realend - realstart),
          file=options.statfile)
    exit(0)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Functions to guess gradation from partial lexical data."""

# Author: Omorfi contributors <omorfi-devel@groups.google.com> 2015

#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.

from .omorfi.error_logging import fail_guess_because
from .omorfi.string_mangling import replace_rightmost, replace_rightmosts


def gradation_make_morphophonemes(wordmap):
    '''mark up gradating stop for morphophonological handling'''
    if not wordmap['kotus_tn']:
        return wordmap
    tn = wordmap['kotus_tn']
    av = wordmap['kotus_av']
    if not wordmap['kotus_av']:
        return wordmap
    if tn in range(1, 28) or tn in range(29, 32) or tn in range(52, 66) or tn == 76 or tn in [1009, 1010]:
        # strong root stem
        if av == 'A':
            wordmap['gradestem'] = replace_rightmost(
                wordmap['gradestem'], 'k', '{k~~}')
        elif av == 'B':
            wordmap['gradestem'] = replace_rightmost(
                wordmap['gradestem'], 'p', '{p~~}')
        elif av == 'C':
            wordmap['gradestem'] = replace_rightmost(
                wordmap['gradestem'], 't', '{t~~}')
        elif av == 'D':
            wordmap['gradestem'] = replace_rightmost(
                wordmap['gradestem'], 'k', '{k~~}')
        elif av == 'E':
            wordmap['gradestem'] = replace_rightmost(
                wordmap['gradestem'], 'p', '{p~~}')
        elif av == 'F':
            wordmap['gradestem'] = replace_rightmost(
                wordmap['gradestem'], 't', '{t~~}')
        elif av == 'G':
            wordmap['gradestem'] = replace_rightmost(
                wordmap['gradestem'], 'k', '{k~~}')
        elif av == 'H':
            wordmap['gradestem'] = replace_rightmost(
                wordmap['gradestem'], 'p', '{p~~}')
        elif av == 'I':
            wordmap['gradestem'] = replace_rightmost(
                wordmap['gradestem'], 't', '{t~~}')
        elif av == 'J':
            wordmap['gradestem'] = replace_rightmost(
                wordmap['gradestem'], 't', '{t~~}')
        elif av == 'K':
            wordmap['gradestem'] = replace_rightmost(
                wordmap['gradestem'], 't', '{t~~}')
        elif av == 'L':
            wordmap['gradestem'] = replace_rightmost(
                wordmap['gradestem'], 'k', '{k~~}')
        elif av == 'M':
            wordmap['gradestem'] = replace_rightmost(
                wordmap['gradestem'], 'k', '{k~~}')
        elif av == 'O':
            wordmap['gradestem'] = replace_rightmost(
                wordmap['gradestem'], 'g', '{g~~}')
        elif av == 'P':
            wordmap['gradestem'] = replace_rightmost(
                wordmap['gradestem'], 'b', '{b~~}')
        elif av == 'N':
            wordmap['gradestem'] = replace_rightmost(
                wordmap['gradestem'], 'd', '{d~~}')
        else:
            fail_guess_because(wordmap, ["1-27", "29-32", "52-66", "76"],
                               ["A-P"])
            exit(1)
        return wordmap
    elif tn == 28:
        # gah gradation in stemparts
        return wordmap
    elif tn == 1007:
        # not gradation
        return wordmap
    else:
        # weak root stem
        if av == 'A':
            wordmap['gradestem'] = replace_rightmost(
                wordmap['gradestem'], 'k', 'k{k~~}')
        elif av == 'B':
            wordmap['gradestem'] = replace_rightmost(
                wordmap['gradestem'], 'p', 'p{p~~}')
        elif av == 'C' and tn in range(72, 76):
            # TVtA verbs
            wordmap['gradestem'] = wordmap['gradestem'][:wordmap['gradestem'].rfind(
                't') - 2] + 't{t~~}' + wordmap['gradestem'][wordmap['gradestem'].rfind('t') - 1:]
        elif av == 'C':
            wordmap['gradestem'] = replace_rightmost(
                wordmap['gradestem'], 't', 't{t~~}')
        elif av == 'D':
            # danger ahead! (it appears from middle of nowhere)
            if tn == 32:
                wordmap['gradestem'] = replace_rightmost(
                    wordmap['gradestem'], 'e', '{k~~}e')
            elif tn == 33:
                wordmap['gradestem'] = replace_rightmost(
                    wordmap['gradestem'], 'i', '{k~~}i')
            elif tn == 41:
                wordmap['gradestem'] = replace_rightmosts(wordmap['gradestem'], [
                                                          'as', 'es', 'is', '{aä}s'], ['{k~~}as', '{k~0}es', '{k~0}is', '{k~~}{aä}s'])
            elif tn == 43:
                wordmap['gradestem'] = replace_rightmosts(
                    wordmap['gradestem'], ['ut', 'yt'], ['{k~~}ut', '{k~~}yt'])
            elif tn == 44:
                wordmap['gradestem'] = replace_rightmost(
                    wordmap['gradestem'], 'e', '{k~~}e')
            elif tn == 48:
                wordmap['gradestem'] = replace_rightmost(
                    wordmap['gradestem'], 'e', '{k~~}e')
            elif tn == 49:
                wordmap['gradestem'] = replace_rightmost(
                    wordmap['gradestem'], 'en', '{k~~}en')
            elif tn == 67:
                wordmap['gradestem'] = replace_rightmost(
                    wordmap['gradestem'], 'ell', '{k~~}ell')
            elif tn == 72:
                wordmap['gradestem'] = replace_rightmosts(
                    wordmap['gradestem'], ['at', 'et', 'ot'], ['{k~~}at', '{k~~}et', '{k~0}ot'])
            elif tn == 73:
                wordmap['gradestem'] = replace_rightmosts(
                    wordmap['gradestem'], ['ata', 'ätä'], ['{k~~}ata', '{k~0}ätä'])
            elif tn == 74:
                wordmap['gradestem'] = replace_rightmosts(wordmap['gradestem'], [
                                                          'ota', 'eta', 'ötä', 'etä'], ['{k~~}ota', '{k~0}eta', '{k~0}ötä', '{k~0}etä'])
            elif tn == 75:
                wordmap['gradestem'] = replace_rightmosts(
                    wordmap['gradestem'], ['ita', 'itä'], ['{k~~}ita', '{k~0}itä'])
            elif tn == 99:
                pass
            else:
                fail_guess_because(wordmap, ["1-99"], ["A-C", "E-P"])
                exit(1)
        elif av == 'E':
            wordmap['gradestem'] = replace_rightmost(
                wordmap['gradestem'], 'v', '{p~~}')
        elif av == 'F':
            wordmap['gradestem'] = replace_rightmost(
                wordmap['gradestem'], 'd', '{t~~}')
        elif av == 'G':
            wordmap['gradestem'] = replace_rightmost(
                wordmap['gradestem'], 'g', '{k~~}')
        elif av == 'H':
            wordmap['gradestem'] = replace_rightmost(
                wordmap['gradestem'], 'm', '{p~~}')
        elif av == 'I' and tn != 67:
            wordmap['gradestem'] = replace_rightmost(
                wordmap['gradestem'], 'l', '{t~~}')
        elif av == 'I' and tn == 67:
            wordmap['gradestem'] = replace_rightmost(
                wordmap['gradestem'], 'lell', '{t~~}ell')
        elif av == 'J' and tn != 33:
            wordmap['gradestem'] = replace_rightmost(
                wordmap['gradestem'], 'n', '{t~~}')
        elif av == 'J' and tn == 33:
            wordmap['gradestem'] = replace_rightmost(
                wordmap['gradestem'], 'nin', '{t~~}in')
        elif av == 'K':
            wordmap['gradestem'] = replace_rightmost(
                wordmap['gradestem'], 'r', '{t~~}')
        elif av == 'L':
            wordmap['gradestem'] = replace_rightmost(
                wordmap['gradestem'], 'j', '{k~~}')
        elif av == 'N':
            wordmap['gradestem'] = replace_rightmost(
                wordmap['gradestem'], 'd', 'd{d~~}')
        elif av == 'O':
            wordmap['gradestem'] = replace_rightmost(
                wordmap['gradestem'], 'g', 'g{g~~}')
        elif av == 'P':
            wordmap['gradestem'] = replace_rightmost(
                wordmap['gradestem'], 'b', 'b{b~~}')
        elif av == 'T' and tn == 49:
            wordmap['gradestem'] = replace_rightmost(
                wordmap['gradestem'], 'e', '{t~~}e')
        else:
            fail_guess_because(wordmap, [""], ["1-99", "A-P"])
            exit(1)
        return wordmap
    return None



#!/usr/bin/env python
# -*- coding: utf-8 -*-


"""
Beta-testing a wack idea of not disambiguating and dependency parsing at the
same time. Pay no attention to the man behind the curtains and move along.
"""

# stuff
import xml.etree.ElementTree
from xml.etree.ElementTree import Element

from .evidence import Evidence
from .matcher import Matcher


class Disamparsulator:
    """Disamparsulator is one such non-parser."""

    def __init__(self):
        self.rules = list()

    def frobblesnizz(self, f):
        '''parse disampursalations from XML file.'''
        xmltree = xml.etree.ElementTree.parse(f)
        root = xmltree.getroot()
        if root.get("version") != "0.0.0":
            print("Unsupported version", root.get("version"))
        rules = list()
        for child in root:
            # parse stuff
            if child.tag == 'evidences':
                self.parse_evidences(child)
            else:
                print("Unknown element disamparsulations:", child.tag)
                exit(2)
        return rules

    def parse_evidences(self, evidences: Element):
        for child in evidences:
            if child.tag == 'evidence':
                e = self.parse_evidence(child)
                self.rules.append(e)
            else:
                print("Unknown element in evidences:", child.tag)
                exit(2)

    def parse_evidence(self, evidence: Element):
        '''Parse evidence element block.'''
        e = Evidence()
        e.name = evidence.get('name')
        for child in evidence:
            if child.tag == 'target':
                e.target = self.parse_target(child)
            elif child.tag == 'likelihood':
                e.unlikelihood = -self.parse_likelihood(child)
            elif child.tag == 'depname':
                e.depname = self.parse_depname(child)
            elif child.tag == 'context':
                e.context = self.parse_context(child)
            elif child.tag == 'documentation':
                pass
            else:
                print("Unknown element under evidence:", child.tag)
                exit(2)
        return e

    def parse_target(self, target: Element):
        '''Parse target definition element.'''
        m = Matcher()
        for child in target:
            if child.tag == 'match':
                m = self.parse_match(child)
            else:
                print("Unknown element under target:", child.tag)
                exit(2)
        return m

    def parse_match(self, match: Element):
        m = Matcher()
        for child in match:
            if child.tag == 'upos':
                upos = self.parse_upos(child)
                m.uposes.append(upos)
            elif child.tag == 'lemma':
                ufeats = self.parse_lemma(child)
                m.lemmas.append(ufeats)
            elif child.tag == 'ufeats':
                ufeats = self.parse_ufeats(child)
                m.ufeatses.append(ufeats)
            else:
                print("Unknown element under target:",
                      xml.etree.ElementTree.tostring(child))
        return m

    def parse_likelihood(self, likelihood: Element):
        '''Parse likelihood element.'''
        if likelihood.text == 'usually':
            return 16.0
        elif likelihood.text == 'probably':
            return 8.0
        elif likelihood.text == 'possibly':
            return 2.0
        elif likelihood.text == 'meh':
            return 0.001
        elif likelihood.text == 'unlikely':
            return -16.0
        elif likelihood.text == 'uncommonly':
            return -4.0
        else:
            print("Unknown likelihood:", likelihood.text)
            exit(2)

    def parse_depname(self, depname: Element):
        '''Parse depname element.'''
        return depname.text

    def parse_context(self, context: Element):
        '''Parse context element.'''
        c = dict()
        if context.get("negated"):
            c["negated"] = True
        for child in context:
            if child.tag == 'location':
                c['location'] = self.parse_location(child)
            elif child.tag == 'match':
                c['matcher'] = self.parse_match(child)
            elif child.tag == 'barrier':
                c['barrier'] = self.parse_barrier(child)
            else:
                print("Unknown child for context", child)
                exit(2)
        return c

    def parse_upos(self, upos: Element):
        '''Parse upos element.'''
        if upos.text not in ['NOUN', 'VERB', 'ADV', 'ADJ', 'ADP',
                             'INTJ', 'PUNCT', 'SYM', 'CCONJ', 'SCONJ',
                             'PRON', 'NUM', 'PROPN', 'DET', 'AUX']:
            print("invalid upos in", xml.etree.ElementTree.tostring(upos))
        return upos.text

    def parse_lemma(self, lemma: Element):
        '''Parse upos element.'''
        return lemma.text

    def parse_location(self, location: Element):
        '''Parse location element.'''
        return location.text

    def parse_ufeats(self, ufeats: Element):
        '''Parse ufeats element.'''
        ufs = dict()
        for child in ufeats:
            if child.tag == 'ufeat':
                name, value = self.parse_ufeat(child)
                if not value:
                    ufs[name] = "*AGREEMENT*"
                else:
                    ufs[name] = value
            else:
                print("Unknown child for ufeats", child)
                exit(2)
        return ufs

    def parse_ufeat(self, ufeat: Element):
        '''Parse ufeat element.'''
        return ufeat.get('name'), ufeat.text

    def parse_barrier(self, barrier: Element):
        '''Parse barrier element.'''
        m = Matcher()
        for child in barrier:
            if child.tag == 'match':
                m = self.parse_match(child)
        return m

    def linguisticate(self, sentence: list):
        '''Not a parsing function.'''
        # for each token for each rule apply
        for token in sentence:
            if token.nontoken:
                continue
            for rule in self.rules:
                rule.apply(token, sentence)
            # some things can be pruned
            cleanups = list()
            for analysis in token.analyses:
                if not analysis.udepname:
                    analysis.weight += 500
                    cleanups.append(analysis)
            if len(cleanups) < len(token.analyses):
                for cleanup in cleanups:
                    for analysis in token.analyses:
                        if analysis.upos == cleanup.upos and \
                                analysis.ufeats == cleanup.ufeats and \
                                analysis.udepname:
                            token.analyses.remove(cleanup)
                            break
        # we have to pull out multiple roots
        toproot = None
        minweight = float('inf')
        for token in sentence:
            if token.nontoken:
                continue
            analysis = token.get_best()
            if analysis.udepname == 'root':
                if analysis.weight < minweight:
                    toproot = analysis
                    minweight = analysis.weight
        for token in sentence:
            if token.nontoken:
                continue
            analysis = token.get_best()
            if analysis.udepname == 'root':
                if analysis != toproot:
                    analysis.weight += 500


def main():
    """Invoke a simple CLI analyser."""
    exit(1)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""A command-line interface for omorfi with CONLL-U I/O formats."""

# string munging
from argparse import ArgumentParser, FileType
# CLI stuff
from sys import stderr, stdin, stdout
# statistics
from time import perf_counter, process_time

# omorfi
from omorfi import Omorfi
from omorfi.formats.fileformats import next_conllu
from omorfi.disamparsulate import Disamparsulator


def get_reference_conllu_list(token):
    if not token.gold:
        print("Oracle data missing from", token, file=stderr)
        exit(2)
    return token.gold.split("\t")


def try_analyses_conllu(token, outfile, hacks=None):
    anals = token.analyses
    original = get_reference_conllu_list(token)
    best = None
    highest = -1
    for i, anal in enumerate(anals):
        upos = anal.get_upos()
        feats = anal.printable_ud_feats()
        lemmas = anal.get_lemmas()
        dephead = anal.udeppos
        depname = anal.udepname
        if lemmas:
            lemma = '#'.join(anal.get_lemmas())
        else:
            lemma = '_'
        score = 0
        if upos == original[3]:
            score += 10
        if lemma == original[2]:
            score += 10
        elif lemma.strip('#') == original[2].strip('#'):
            score += 5
        elif lemma.lower() == original[2].lower():
            score += 5
        if feats == original[5]:
            score += 10
        else:
            featset = set(feats.split("|"))
            refset = set(original[5].split("|"))
            score += len(featset.intersection(refset))
        if dephead == original[6]:
            score += 1
        if depname == original[7]:
            score += 1
        if score > highest:
            best = i
            highest = score
    print(token.printable_conllu(hacks, best), file=outfile)


def debug_analyses_conllu(token, outfile, hacks=None):
    anals = token.analyses
    print("# REFERENCE(python):", get_reference_conllu_list(token),
          file=outfile)
    for i, anal in enumerate(anals):
        print(token.printable_conllu(hacks, i), file=outfile)


def print_analyses(sent, options):
    for token in sent:
        if token.nontoken:
            print(token.printable_conllu(options.hacks), file=options.outfile)
        elif options.debug:
            debug_analyses_conllu(token, options.outfile, options.hacks)
        elif options.oracle:
            try_analyses_conllu(token, options.outfile, options.hacks)
        else:
            print(token.printable_conllu(options.hacks), file=options.outfile)


def main():
    """Invoke a simple CLI analyser."""
    a = ArgumentParser()
    a.add_argument('-a', '--analyser', metavar='AFILE', required=True,
                   help="read analyser model from AFILE")
    a.add_argument('-i', '--input', metavar="INFILE", type=open,
                   dest="infile", help="source of analysis data")
    a.add_argument('-v', '--verbose', action='store_true',
                   help="print verbosely while processing")
    a.add_argument('-o', '--output', metavar="OUTFILE", dest="outfile",
                   help="print output into OUTFILE", type=FileType('w'))
    a.add_argument('-x', '--statistics', metavar="STATFILE", dest="statfile",
                   help="print statistics to STATFILE", type=FileType('w'))
    a.add_argument('-O', '--oracle', action='store_true',
                   help="match to values in input when parsing if possible")
    a.add_argument('-u', '--udpipe', metavar="UDPIPE",
                   help='use UDPIPE for additional guesses (experi-mental)')
    a.add_argument('--hacks', metavar='HACKS',
                   help="mangle analyses to match HACKS version of UD",
                   choices=['ftb'])
    a.add_argument('-X', '--frequencies', metavar="FREQDIR",
                   help="read frequencies from FREQDIR/*.freqs")
    a.add_argument('--not-rules', metavar="RULEFILE", type=open,
                   help="read non-rules from RULEFILE")
    a.add_argument('--debug', action='store_true',
                   help="print lots of debug info while processing")
    options = a.parse_args()
    if options.verbose:
        print("Printing verbosely")
    omorfi = Omorfi(options.verbose)
    if options.analyser:
        if options.verbose:
            print("reading analyser model", options.analyser)
        omorfi.load_analyser(options.analyser)
    else:
        print("analyser is needed to conllu", file=stderr)
        exit(4)
    disamparsulator = None
    if options.not_rules:
        disamparsulator = Disamparsulator()
        if options.verbose:
            print("Loading", options.not_rules)
        disamparsulator.frobblesnizz(options.not_rules)
    if options.udpipe:
        if options.verbose:
            print("Loading udpipe", options.udpipe)
        omorfi.load_udpipe(options.udpipe)
    if not options.infile:
        print("reading from <stdin>")
        options.infile = stdin
    if options.verbose:
        print("analysing", options.infile.name)
    if not options.outfile:
        options.outfile = stdout
    if options.verbose:
        print("writing to", options.outfile.name)
    if not options.statfile:
        options.statfile = stdout

    if options.frequencies:
        with open(options.frequencies + '/lexemes.freqs') as lexfile:
            omorfi.load_lexical_frequencies(lexfile)
        with open(options.frequencies + '/omors.freqs') as omorfile:
            omorfi.load_omortag_frequencies(omorfile)

    # statistics
    realstart = perf_counter()
    cpustart = process_time()
    tokens = 0
    unknowns = 0
    sentences = 0
    eoffed = False
    while not eoffed:
        sentplus = next_conllu(options.infile)
        if not sentplus:
            eoffed = True
            break
        for token in sentplus:
            if token.nontoken:
                if token.nontoken == 'comment':
                    pass
                elif token.nontoken == 'eof':
                    eoffed = True
                    break
                elif token.nontoken == 'separator':
                    sentences += 1
                elif token.nontoken == 'error':
                    print("Unrecognisable line:", token.error, file=stderr)
                    exit(1)
                else:
                    print("Error:", token, file=stderr)
                    exit(1)
                continue
            elif not token.surf:
                print("No surface in CONLL-U?", token, file=stderr)
                exit(1)
            tokens += 1
            omorfi.analyse(token)
            if token.is_oov():
                unknowns += 1
                omorfi.guess(token)
        if disamparsulator:
            disamparsulator.linguisticate(sentplus)
        print_analyses(sentplus, options)
    cpuend = process_time()
    realend = perf_counter()
    print("Tokens:", tokens, "Sentences:", sentences,
          file=options.statfile)
    print("Unknowns / OOV:", unknowns, "=",
          unknowns / tokens * 100 if tokens != 0 else 0,
          "%", file=options.statfile)
    print("CPU time:", cpuend - cpustart, "Real time:", realend - realstart,
          file=options.statfile)
    print("Tokens per timeunit:", tokens / (realend - realstart),
          file=options.statfile)
    exit(0)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Test omorfi - unimorph compatibility
"""


from argparse import ArgumentParser, FileType
from sys import stderr, stdout, stdin
from time import perf_counter, process_time

from omorfi import Omorfi, Token


def main():
    """Command-line interface for omorfi's sort | uniq -c tester."""
    a = ArgumentParser()
    a.add_argument('-a', '--analyser', metavar='FSAFILE', required=True,
                   help="load analyser from FSAFILE")
    a.add_argument('-i', '--input', metavar="INFILE", type=open,
                   dest="infile", help="source of analysis data")
    a.add_argument('-o', '--output', metavar="OUTFILE",
                   type=FileType('w'),
                   dest="outfile", help="log outputs to OUTFILE")
    a.add_argument('-X', '--statistics', metavar="STATFILE",
                   type=FileType('w'),
                   dest="statfile", help="statistics")
    a.add_argument('-v', '--verbose', action="store_true", default=False,
                   help="Print verbosely while processing")
    a.add_argument('-C', '--no-casing', action="store_true", default=False,
                   help="Do not try to recase input and output when matching")
    a.add_argument('-t', '--threshold', metavar="THOLD", default=99,
                   help="if coverage is less than THOLD exit with error")
    options = a.parse_args()
    omorfi = Omorfi(options.verbose)
    try:
        if options.analyser:
            if options.verbose:
                print("reading analyser from", options.analyser)
            omorfi.load_analyser(options.analyser)
        if not options.infile:
            options.infile = stdin
            print("reading from <stdin>")
        if not options.statfile:
            options.statfile = stdout
        if not options.outfile:
            options.outfile = stdout
    except IOError:
        print("Could not process file", options.analyser, file=stderr)
        exit(2)
    # basic statistics
    covered = 0
    full_matches = 0
    lemma_matches = 0
    anal_matches = 0
    no_matches = 0
    no_results = 0
    only_permuted = 0
    accfails = 0
    lines = 0
    # for make check target
    threshold = options.threshold
    realstart = perf_counter()
    cpustart = process_time()
    for line in options.infile:
        fields = line.strip().split('\t')
        if len(fields) < 3:
            print("ERROR: Skipping line", fields, file=stderr)
            continue
        if ' ' in fields[1] or ' ' in fields[0]:
            continue
        lines += 1
        if options.verbose and lines % 1000 == 0:
            print(lines, '...')
        lemma = fields[0]
        surf = fields[1]
        unimorph = fields[2].replace('ACC', 'NOM').replace('GEADJ', 'GEN')
        token = Token(surf)
        omorfi.analyse(token)
        if not token.is_oov():
            covered += 1
        else:
            no_results += 1
            print(1, "OOV", surf, sep='\t', file=options.outfile)
        found_anals = False
        found_lemma = False
        permuted = True
        accfail = False
        for anal in token.analyses:
            analhyp = anal.printable_unimorph()
            lemmahyp = ''.join(anal.get_lemmas())
            if analhyp == unimorph:
                found_anals = True
                permuted = False
            elif set(analhyp.split(';')) == set(unimorph.split(';')):
                found_anals = True
                # print("PERMUTAHIT", analhyp, unimorph, sep='\t',
                #      file=options.outfile)
            else:
                pass
                # print("ANALMISS", analhyp, unimorph, sep='\t',
                #      file=options.outfile)
            if lemma == lemmahyp:
                found_lemma = True
            else:
                pass
                # print("LEMMAMISS", lemmahyp, lemma, sep='\t',
                #      file=options.outfile)
        if not found_anals and not found_lemma:
            no_matches += 1
            print("NOHITS!", surf, lemma, unimorph,
                  [a.printable_unimorph() for a in token.analyses], sep='\t',
                  file=options.outfile)
        elif found_anals and found_lemma:
            full_matches += 1
        elif not found_anals:
            anal_matches += 1
            print("LEMMANOANAL", surf, unimorph,
                  [a.printable_unimorph() for a in token.analyses], sep='\t',
                  file=options.outfile)
        elif not found_lemma:
            lemma_matches += 1
            print("ANALNOLEMMA", surf, lemma,
                  [a.get_lemmas() for a in token.analyses], sep='\t',
                  file=options.outfile)
        else:
            print("Logical error, kill everyone")
            exit(13)
        if permuted:
            only_permuted += 1
        if accfail:
            accfails += 1
    realend = perf_counter()
    cpuend = process_time()
    print("CPU time:", cpuend - cpustart, "real time:", realend - realstart)
    if lines == 0:
        print("Needs more than 0 lines to determine something",
              file=stderr)
        exit(2)
    print("Lines", "Covered", "OOV", sep="\t", file=options.statfile)
    print(lines, covered, lines - covered, sep="\t", file=options.statfile)
    print(lines / lines * 100 if lines != 0 else 0,
          covered / lines * 100 if lines != 0 else 0,
          (lines - covered) / lines * 100 if lines != 0 else 0,
          sep="\t", file=options.statfile)
    print("Lines", "Matches", "Lemma", "Anals", "Mismatch",
          "No results", sep="\t", file=options.statfile)
    print(lines, full_matches, lemma_matches, anal_matches, no_matches,
          no_results,
          sep="\t", file=options.statfile)
    print(lines / lines * 100 if lines != 0 else 0,
          full_matches / lines * 100 if lines != 0 else 0,
          lemma_matches / lines * 100 if lines != 0 else 0,
          anal_matches / lines * 100 if lines != 0 else 0,
          no_matches / lines * 100 if lines != 0 else 0,
          no_results / lines * 100 if lines != 0 else 0,
          sep="% \t", file=options.statfile)
    print("Of which", "Tag permuations", sep='\t',
          file=options.statfile)
    print(lines / lines * 100 if lines != 0 else 0,
          only_permuted / lines * 100 if lines != 0 else 0, sep='\t',
          file=options.statfile)
    if full_matches / lines * 100 <= int(options.threshold):
        print("needs to have", threshold, "% matches to pass regress test\n",
              "please examine", options.outfile.name, "for regressions",
              file=stderr)
        exit(1)
    elif covered / lines * 100 <= int(options.threshold):
        print("needs to have", threshold, "% coverage to pass regress test\n",
              "please examine", options.outfile.name, "for regressions",
              file=stderr)
        exit(1)
    else:
        exit(0)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Support functions for handling tokens. Now in class form.
"""
# debug :-(
from sys import stderr

from .analysis import Analysis


class Token:
    """Token holds a slice of text with its analyses and features.

    Token is typically a word-form, such as "pojat" or "juopottelevat", but
    can also be a white-space sequence or placeholder for some out of text
    metadata, a tag, comment or i/o error.

    For a reference, see [spaCy tokens](https://spacy.io/api/token), it's
    not exactly the same thing and I don't agree with all there, but it's
    quite cool and well-documented.
    """

    def __init__(self, surf=None):
        """Create token with surface string optionally."""
        ## all gathered analyses so far as Analysis objects
        self.analyses = []
        ## all constructed morph segmentations
        self.segmentations = []
        ## all constructed morph labelsegmentations
        self.labelsegmentations = []
        ## all constructed lemmatisations
        self.lemmatisations = []
        ## original surface form
        self.surf = surf
        ## word index in context, e.g. UD column 1
        self.pos = 0
        ## nontoken
        self.nontoken = False
        ## comment (esp. with non-token)
        self.comment = None
        ## use when tokenisation or parsing breaks
        self.error = None
        ## If token is separated by space from left
        self.spacebefore = False
        ## If token is separated by space from right
        self.spaceafter = False
        ## Gold reference can be stored in token for few apps
        self.gold = None

    def __getitem__(self, key):
        """Tokens can still be accessed like dicts for compatibility.

        Some keys like surf and pos are obvious and direct while some old keys
        like omor for analysis is mapped to 1-random analysis string if there
        are any.
        """
        if key == 'anal':
            # return first omor analysis for b/w comp & ease of use
            for analysis in self.analyses:

                if analysis.name == 'omor':
                    return analysis.raw
            return None
        elif key == 'surf':
            return self.surf
        elif key == 'pos':
            return self.pos
        else:
            raise KeyError(key)

    def __str__(self):
        s = '"omorfi.Token": {'
        if self.surf:
            s += '"surf": "' + self.surf + '"'
        if self.nontoken:
            s += '"nontoken": "' + self.nontoken + '"'
        if self.pos:
            s += ',\n "pos": "' + str(self.pos) + '"'
        if self.analyses:
            s += ',\n "omorfi.Analyses": ['
            comma = ""
            for anal in self.analyses:
                s += comma + "\n" + str(anal)
                comma = ","
            s += '\n]'
        if self.error:
            s += ',\n "error": "' + self.error + '"'
        if self.comment:
            s += ',\n "comment": "' + self.comment + '"'
        s += '}'
        return s

    @staticmethod
    def fromstr(s: str):
        """Creates token from string.

        Strings should be made with print(token).
        """
        lines = s.split("\n")
        token = Token()
        in_anals = False
        for line in lines:
            if '"omorfi.Token"' in line:
                # first
                continue
            elif line.strip() == '}':
                # last
                continue
            elif in_anals:
                if line.strip() == ']':
                    in_anals = False
                else:
                    anal = Analysis.fromstr(line)
                    token.analyses.append(anal)
            else:
                k, v = line.split(":")
                if k == '"surf"':
                    token.surf = v.strip().strip('"')
                elif k == '"nontoken"':
                    token.nontoken = v.strip().strip('"')
                elif k == '"pos"':
                    token.pos = int(v.strip().strip('"'))
                elif k == '"error"':
                    token.error = v.strip().strip('"')
                elif k == '"comment"':
                    token.comment = v.strip().strip('"')
                elif k == '"omorfi.Analyses"':
                    in_anals = True
                else:
                    print("Error parsing token", line, file=stderr)
                    exit(1)
        return token

    @staticmethod
    def fromdict(token: dict):
        """Create token from pre-2019 tokendict."""
        cons = Token(token['surf'])
        for k, v in token.items():
            if k == 'anal':
                anal = Analysis(v, 0.0)
                cons.analyses.append(anal)
        return cons

    @staticmethod
    def fromsurf(surf: str):
        """Creat token from surface string."""
        return Token(surf)

    @staticmethod
    def fromconllu(conllu: str):
        """Create token from conll-u line."""
        fields = conllu.split()
        if len(fields) != 10:
            print("conllu2token conllu fail", fields)
        upos = fields[3]
        wordid = fields[2]
        surf = fields[1]
        ufeats = fields[5]
        misc = fields[9]
        omor = Analysis('[WORD_ID=' + wordid + ']' + '[UPOS=' + upos + ']' +
                        Token._ufeats2omor(ufeats), 0.0, "omor")
        omor.misc = misc
        token = Token(surf)
        token.analyses.append(omor)
        # we can now store the original here
        orig = Analysis(conllu, 0.0, "conllu")
        token.analyses.append(orig)
        token.surf = surf
        return token

    @staticmethod
    def fromvislcg(s: str):
        '''Create a token from VISL CG-3 text block.

        The content should at most contain one surface form with a set of
        analyses.
        '''
        token = Token()
        lines = s.split('\n')
        for line in lines:
            line = line.rstrip()
            if not line or line == '':
                token.nontoken = 'separator'
            elif line.startswith("#") or line.startswith("<"):
                token.nontoken = 'comment'
                token.comment = line.strip()
            elif line.startswith('"<') and line.endswith('>"'):
                token.surf = line[2:-2]
            elif line.startswith('\t"'):
                anal = Analysis(line.strip(), 0.0, "vislcg")
                token.analyses.append(anal)
            elif line.startswith(';\t"'):
                # gold?
                anal = Analysis(line[1:].strip(), float("inf"), "vislcg")
                token.analyses.append()
            else:
                token.nontoken = "error"
                token.error = 'vislcg: ' + line.strip()
        return token

    @staticmethod
    def _ufeats2omor(ufeats):
        return '[' + ufeats.replace('|', '][') + ']'

    def is_oov(self):
        '''Checks if all hypotheses are OOV guesses.'''
        for analysis in self.analyses:
            if not analysis.is_oov():
                return False
        return True

    def printable_vislcg(self):
        '''Create VISL-CG 3 output based on token and its analyses.'''
        if self.error:
            return "<ERROR>" + self.error + "</ERROR>"
        elif self.nontoken:
            if self.nontoken == 'comment':
                return self.comment
            elif self.nontoken == 'separator':
                return '\n'
            elif self.nontoken == 'eof':
                return ""
            else:
                return "<ERROR>" + self.nontoken + "</ERROR>"
        elif self.surf:
            vislcg = '"<' + self.surf + '>"\n'
            newline = ""
            for anal in self.analyses:
                vislcg += newline + anal.printable_vislcg()
                newline = "\n"
            return vislcg
        else:
            return "<ERROR>" + self + "</ERROR>"

    def _select_anal(self, which):
        anal = None
        if which == "1random":
            if self.analyses:
                anal = self.analyses[0]
        elif which == "1best":
            minw = float("inf")
            for a in self.analyses:
                if a.weight < minw:
                    anal = a
                    minw = a.weight
        elif self.analyses[which]:
            anal = self.analyses[which]
        else:
            print("Unknown which", which)
            exit(1)
        return anal

    def printable_conllu(self, hacks=None, which="1best"):
        '''Create CONLL-U output based on token and selected analysis.'''
        if self.nontoken:
            if self.nontoken == 'error':
                return "# ERROR:" + self.error
            elif self.nontoken == 'comment':
                if self.comment.startswith('#'):
                    return self.comment
                else:
                    return '# ' + self.comment
            elif self.nontoken == 'separator':
                # not returning \n since the it's already printed on a line
                return ''
            else:
                # ignore other nontokens??
                return ''
        lemma = self.surf
        upos = 'X'
        third = '_'
        ud_feats = '_'
        ud_misc = '_'
        anal = self._select_anal(which)
        if anal:
            upos = anal.get_upos()
            if hacks and hacks == 'ftb':
                third = anal.get_xpos_ftb()
            else:
                third = anal.get_xpos_tdt()
            lemmas = anal.get_lemmas()
            if lemmas:
                lemma = '#'.join(lemmas)
            ud_feats = anal.printable_ud_feats()
            ud_misc = anal.printable_ud_misc()
            depname = anal.printable_udepname()
            dephead = anal.printable_udephead()
        return "\t".join([str(self.pos), self.surf, lemma, upos, third,
                          ud_feats, dephead, depname, "_", ud_misc])

    def printable_ftb3(self, which="1best"):
        '''Create FTB-3 output based on token and selected analysis.'''
        if self.nontoken:
            if self.nontoken == 'error':
                return "# ERROR:" + self.error
            elif self.nontoken == 'comment':
                if self.comment.startswith('#'):
                    return self.comment
                else:
                    return '# ' + self.comment
            elif self.nontoken == 'separator':
                # not returning \n since the it's already printed on a line
                return ''
            else:
                # ignore other nontokens??
                return ''
        lemma = self.surf
        pos = '_'
        feats = '_'
        anal = self._select_anal(which)
        if anal:
            pos = anal.get_xpos_ftb()
            lemmas = anal.get_lemmas()
            if lemmas:
                lemma = '#'.join(lemmas)
            feats = anal.printable_ftb_feats()
        return '\t'.join([str(self.pos), self.surf, lemma, pos, pos, feats,
                          '_', '_', '_', '_'])

    def _get_nbest(self, n: int, anals: list):
        nbest = []
        worst = -1.0
        if n == 0:
            n = 65535
        for anal in anals:
            if len(nbest) < n:
                nbest.append(anal)
                # when filling the queue find biggest
                if anal.weight > worst:
                    worst = anal.weight
            elif anal.weight < worst:
                # replace worst
                for i, a in enumerate(nbest):
                    if a.weight == worst:
                        nbest[i] = anal
                worst = -1.0
                for a in nbest:
                    if a.weight > worst:
                        worst = a.weight
        return nbest

    def get_nbest(self, n: int):
        """Get n most likely analyses.

        Args:
            n: number of analyses, use 0 to get all

        Returns:
            At most n analyses of given type or empty list if there aren't any.
        """
        return self._get_nbest(n, self.analyses)

    def get_best(self):
        """Get most likely analysis.

        Returns:
            most probably analysis of given type, or None if analyses have not
            been made for the type.
        """
        nbest1 = self.get_nbest(1)
        if nbest1:
            return nbest1[0]
        else:
            return None

    def get_best_segments(self):
        """Get most likely segmentation.

        Returns:
            list of strings each one being a morph or other sub-word segment.
        """
        nbest1 = self._get_nbest(1, self.segmentations)
        if nbest1:
            return nbest1[0]
        else:
            return None



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Create wiktionary-like inflection tables for omorfi paradigms and words.
"""


from argparse import ArgumentParser, FileType
from sys import stderr, stdout
from time import perf_counter, process_time

from omorfi import Omorfi


def generate(generator, word, upos, omors=""):
    """Generate word-form."""
    wf = generator.generate("[WORD_ID=" + word + "][UPOS=" + upos + "]" +
                            omors)
    if not wf:
        return '---'
    else:
        return wf


def print_nominals_(omorfi, word, upos, outfile, inject=''):
    print("| *Case* | _Singular_ | _Plural_ | _Possessive_ | _Clitic_ |",
          file=outfile)
    print("|:-------|:-----------|:---------|:-------------|:---------|",
          file=outfile)
    print("| _Nominative_ | " +
          generate(omorfi, word, upos, inject +
                   "[NUM=SG][CASE=NOM]") + " | " +
          generate(omorfi, word, upos, inject +
                   "[NUM=PL][CASE=NOM]") + " | " +
          generate(omorfi, word, upos, inject +
                   "[NUM=SG][CASE=NOM][POSS=SG1]") + " | " +
          generate(omorfi, word, upos, inject +
                   "[NUM=PL][CASE=NOM][CLIT=HAN]") + " |",
          file=outfile)
    print("| _Genitive_ | " +
          generate(omorfi, word, upos, inject +
                   "[NUM=SG][CASE=GEN]") + " | " +
          generate(omorfi, word, upos, inject +
                   "[NUM=PL][CASE=GEN]") + " | " +
          generate(omorfi, word, upos, inject +
                   "[NUM=SG][CASE=GEN][POSS=SG2]") + " | " +
          generate(omorfi, word, upos, inject +
                   "[NUM=PL][CASE=GEN][CLIT=KO]") + " |",
          file=outfile)
    print("| _Partitive_ | " +
          generate(omorfi, word, upos, inject +
                   "[NUM=SG][CASE=PAR]") + " | " +
          generate(omorfi, word, upos, inject +
                   "[NUM=PL][CASE=PAR]") + " | " +
          generate(omorfi, word, upos, inject +
                   "[NUM=SG][CASE=PAR][POSS=3]") + " | " +
          generate(omorfi, word, upos, inject +
                   "[NUM=PL][CASE=PAR][CLIT=PA]") + " |",
          file=outfile)
    print("| _Inessive_ | " +
          generate(omorfi, word, upos, inject +
                   "[NUM=SG][CASE=INE]") + " | " +
          generate(omorfi, word, upos, inject +
                   "[NUM=PL][CASE=INE]") + " | " +
          generate(omorfi, word, upos, inject +
                   "[NUM=SG][CASE=INE][POSS=PL1]") + " | " +
          generate(omorfi, word, upos, inject +
                   "[NUM=PL][CASE=INE][CLIT=KAAN]") + " |",
          file=outfile)
    print("| _Elative_ | " +
          generate(omorfi, word, upos, inject +
                   "[NUM=SG][CASE=ELA]") + " | " +
          generate(omorfi, word, upos, inject +
                   "[NUM=PL][CASE=ELA]") + " | " +
          generate(omorfi, word, upos, inject +
                   "[NUM=SG][CASE=ELA][POSS=PL2]") + " | " +
          generate(omorfi, word, upos, inject +
                   "[NUM=PL][CASE=ELA][CLIT=KIN]") + " |",
          file=outfile)
    print("| _Illative_ | " +
          generate(omorfi, word, upos, inject +
                   "[NUM=SG][CASE=ILL]") + " | " +
          generate(omorfi, word, upos, inject +
                   "[NUM=PL][CASE=ILL]") + " | " +
          generate(omorfi, word, upos, inject +
                   "[NUM=SG][CASE=ILL][POSS=3]") + " | " +
          generate(omorfi, word, upos, inject +
                   "[NUM=PL][CASE=ILL][CLIT=KO]") + " |",
          file=outfile)
    print("| _Adessive_ | " +
          generate(omorfi, word, upos, inject +
                   "[NUM=SG][CASE=ADE]") + " | " +
          generate(omorfi, word, upos, inject +
                   "[NUM=PL][CASE=ADE]") + " | " +
          generate(omorfi, word, upos, inject +
                   "[NUM=SG][CASE=ADE][POSS=SG1]") + " | " +
          generate(omorfi, word, upos, inject +
                   "[NUM=PL][CASE=ADE][CLIT=PA]") + " |",
          file=outfile)
    print("| _Ablative_ | " +
          generate(omorfi, word, upos, inject +
                   "[NUM=SG][CASE=ABL]") + " | " +
          generate(omorfi, word, upos, inject +
                   "[NUM=PL][CASE=ABL]") + " | " +
          generate(omorfi, word, upos, inject +
                   "[NUM=SG][CASE=ABL][POSS=SG2]") + " | " +
          generate(omorfi, word, upos, inject +
                   "[NUM=PL][CASE=ABL][CLIT=KAAN]") + " |",
          file=outfile)
    print("| _Allative_ | " +
          generate(omorfi, word, upos, inject +
                   "[NUM=SG][CASE=ALL]") + " | " +
          generate(omorfi, word, upos, inject +
                   "[NUM=PL][CASE=ALL]") + " | " +
          generate(omorfi, word, upos, inject +
                   "[NUM=SG][CASE=ALL][POSS=3]") + " | " +
          generate(omorfi, word, upos, inject +
                   "[NUM=PL][CASE=ALL][CLIT=HAN]") + " |",
          file=outfile)
    print("| _Essive_ | " +
          generate(omorfi, word, upos, inject +
                   "[NUM=SG][CASE=ESS]") + " | " +
          generate(omorfi, word, upos, inject +
                   "[NUM=PL][CASE=ESS]") + " | " +
          generate(omorfi, word, upos, inject +
                   "[NUM=SG][CASE=ESS][POSS=PL1]") + " | " +
          generate(omorfi, word, upos, inject +
                   "[NUM=PL][CASE=ESS][CLIT=KO]") + " |",
          file=outfile)
    print("| _Translative_ | " +
          generate(omorfi, word, upos, inject +
                   "[NUM=SG][CASE=TRA]") + " | " +
          generate(omorfi, word, upos, inject +
                   "[NUM=PL][CASE=TRA]") + " | " +
          generate(omorfi, word, upos, inject +
                   "[NUM=SG][CASE=TRA][POSS=PL2]") + " | " +
          generate(omorfi, word, upos, inject +
                   "[NUM=PL][CASE=TRA][CLIT=PA]") + " |",
          file=outfile)
    print("| _Abessive_ | " +
          generate(omorfi, word, upos, inject +
                   "[NUM=SG][CASE=ABE]") + " | " +
          generate(omorfi, word, upos, inject +
                   "[NUM=PL][CASE=ABE]") + " | " +
          generate(omorfi, word, upos, inject +
                   "[NUM=SG][CASE=ABE][POSS=3]") + " | " +
          generate(omorfi, word, upos, inject +
                   "[NUM=PL][CASE=ABE][CLIT=KAAN]") + " |",
          file=outfile)
    print("| _Instructive_ | " +
          generate(omorfi, word, upos, inject +
                   "[NUM=SG][CASE=INS]") + " | " +
          generate(omorfi, word, upos, inject +
                   "[NUM=PL][CASE=INS]") + " | " +
          generate(omorfi, word, upos, inject +
                   "[NUM=SG][CASE=INS][POSS=SG1]") + " | " +
          generate(omorfi, word, upos, inject +
                   "[NUM=PL][CASE=INS][CLIT=KO]") + " |",
          file=outfile)
    print("| _Comitative_ | " +
          generate(omorfi, word, upos, inject +
                   "[CASE=COM]") + " | " +
          generate(omorfi, word, upos, inject +
                   "[CASE=COM]") + " | " +
          generate(omorfi, word, upos, inject +
                   "[CASE=COM][POSS=SG2]") + " | " +
          generate(omorfi, word, upos, inject +
                   "[CASE=COM][CLIT=HAN]") + " |",
          file=outfile)


def print_nominals(omorfi, word, upos, outfile):
    print("#### Nominal cases", file=outfile)
    print(file=outfile)
    print_nominals_(omorfi, word, upos, outfile, '')


def print_numerals(omorfi, word, upos, outfile):
    print("#### Nominal cases", file=outfile)
    print(file=outfile)
    if word.endswith('s') or word.endswith('nen'):
        tags = '[NUMTYPE=ORD]'
    else:
        tags = '[NUMTYPE=CARD]'
    print_nominals_(omorfi, word, upos, outfile, tags)

def print_comparatives(omorfi, word, upos, comp, outfile):
    if comp == 'POS':
        print("#### Nominal cases", file=outfile)
        tags = "[CMP=POS]"
    elif comp == 'CMP':
        print("#### Comparative cases", file=outfile)
        tags = "[DRV=MPI][CMP=CMP]"
    elif comp == 'SUP':
        print("#### Superlative cases", file=outfile)
        tags = "[DRV=IN²][CMP=SUP]"
    print(file=outfile)
    print_nominals_(omorfi, word, upos, outfile, tags)


def print_finites(omorfi, word, upos, outfile):
    print(file=outfile)
    print("#### Finite forms", file=outfile)
    print(file=outfile)
    print("| *Person* | _Present_ | _Past_ | _Conditional_ |" +
          " _Potential_ | _Imperative_ |", file=outfile)
    print("|:---------|:----------|:-------|:--------------|"
          ":------------|:-------------|", file=outfile)
    print("| _1st singular_ | " +
          generate(omorfi, word, upos,
                   "[VOICE=ACT][MOOD=INDV][TENSE=PRESENT][PERS=SG1]") +
          " | " +
          generate(omorfi, word, upos,
                   "[VOICE=ACT][MOOD=INDV][TENSE=PAST][PERS=SG1]") +
          " | " +
          generate(omorfi, word, upos,
                   "[VOICE=ACT][MOOD=COND][PERS=SG1]") +
          " | " +
          generate(omorfi, word, upos,
                   "[VOICE=ACT][MOOD=POTN][PERS=SG1]") +
          " | " +
          generate(omorfi, word, upos,
                   "[VOICE=ACT][MOOD=IMPV][PERS=SG1]") + " |",
          file=outfile)
    print("| _2nd singular_ | " +
          generate(omorfi, word, upos,
                   "[VOICE=ACT][MOOD=INDV][TENSE=PRESENT][PERS=SG2]") +
          " | " +
          generate(omorfi, word, upos,
                   "[VOICE=ACT][MOOD=INDV][TENSE=PAST][PERS=SG2]") +
          " | " +
          generate(omorfi, word, upos,
                   "[VOICE=ACT][MOOD=COND][PERS=SG2]") +
          " | " +
          generate(omorfi, word, upos,
                   "[VOICE=ACT][MOOD=POTN][PERS=SG2]") +
          " | " +
          generate(omorfi, word, upos,
                   "[VOICE=ACT][MOOD=IMPV][PERS=SG2]") + " |",
          file=outfile)
    print("| _3rd singular_ | " +
          generate(omorfi, word, upos,
                   "[VOICE=ACT][MOOD=INDV][TENSE=PRESENT][PERS=SG3]") +
          " | " +
          generate(omorfi, word, upos,
                   "[VOICE=ACT][MOOD=INDV][TENSE=PAST][PERS=SG3]") +
          " | " +
          generate(omorfi, word, upos,
                   "[VOICE=ACT][MOOD=COND][PERS=SG3]") +
          " | " +
          generate(omorfi, word, upos,
                   "[VOICE=ACT][MOOD=POTN][PERS=SG3]") +
          " | " +
          generate(omorfi, word, upos,
                   "[VOICE=ACT][MOOD=IMPV][PERS=SG3]") + " |",
          file=outfile)
    print("| _1st plural_ | " +
          generate(omorfi, word, upos,
                   "[VOICE=ACT][MOOD=INDV][TENSE=PRESENT][PERS=PL1]") +
          " | " +
          generate(omorfi, word, upos,
                   "[VOICE=ACT][MOOD=INDV][TENSE=PAST][PERS=PL1]") +
          " | " +
          generate(omorfi, word, upos,
                   "[VOICE=ACT][MOOD=COND][PERS=PL1]") +
          " | " +
          generate(omorfi, word, upos,
                   "[VOICE=ACT][MOOD=POTN][PERS=PL1]") +
          " | " +
          generate(omorfi, word, upos,
                   "[VOICE=ACT][MOOD=IMPV][PERS=PL1]") + " |",
          file=outfile)
    print("| _2nd plural_ | " +
          generate(omorfi, word, upos,
                   "[VOICE=ACT][MOOD=INDV][TENSE=PRESENT][PERS=PL2]") +
          " | " +
          generate(omorfi, word, upos,
                   "[VOICE=ACT][MOOD=INDV][TENSE=PAST][PERS=PL2]") +
          " | " +
          generate(omorfi, word, upos,
                   "[VOICE=ACT][MOOD=COND][PERS=PL2]") +
          " | " +
          generate(omorfi, word, upos,
                   "[VOICE=ACT][MOOD=POTN][PERS=PL2]") +
          " | " +
          generate(omorfi, word, upos,
                   "[VOICE=ACT][MOOD=IMPV][PERS=PL2]") + " |",
          file=outfile)
    print("| _3rd plural_ | " +
          generate(omorfi, word, upos,
                   "[VOICE=ACT][MOOD=INDV][TENSE=PRESENT][PERS=PL3]") +
          " | " +
          generate(omorfi, word, upos,
                   "[VOICE=ACT][MOOD=INDV][TENSE=PAST][PERS=PL3]") +
          " | " +
          generate(omorfi, word, upos,
                   "[VOICE=ACT][MOOD=COND][PERS=PL3]") +
          " | " +
          generate(omorfi, word, upos,
                   "[VOICE=ACT][MOOD=POTN][PERS=PL3]") +
          " | " +
          generate(omorfi, word, upos,
                   "[VOICE=ACT][MOOD=IMPV][PERS=PL3]") + " |",
          file=outfile)
    print("| _Impersonal_ | " +
          generate(omorfi, word, upos,
                   "[VOICE=PSS][MOOD=INDV][TENSE=PRESENT][PERS=PE4]") +
          " | " +
          generate(omorfi, word, upos,
                   "[VOICE=PSS][MOOD=INDV][TENSE=PAST][PERS=PE4]") +
          " | " +
          generate(omorfi, word, upos,
                   "[VOICE=PSS][MOOD=COND][PERS=PE4]") +
          " | " +
          generate(omorfi, word, upos,
                   "[VOICE=PSS][MOOD=POTN][PERS=PE4]") +
          " | " +
          generate(omorfi, word, upos,
                   "[VOICE=PSS][MOOD=IMPV][PERS=PE4]") + " |",
          file=outfile)
    print("| _Connegative singular_ | " +
          generate(omorfi, word, upos,
                   "[VOICE=ACT][MOOD=INDV][TENSE=PRESENT][NEG=CON]") +
          " | " +
          generate(omorfi, word, upos,
                   "[VOICE=ACT][MOOD=INDV][TENSE=PAST][NUM=SG][NEG=CON]") +
          " | " +
          generate(omorfi, word, upos,
                   "[VOICE=ACT][MOOD=COND][NEG=CON]") +
          " | " +
          generate(omorfi, word, upos,
                   "[VOICE=ACT][MOOD=POTN][NEG=CON]") +
          " | " +
          generate(omorfi, word, upos,
                   "[VOICE=ACT][MOOD=IMPV][NEG=CON]") + " |",
          file=outfile)
    print("| _Connegative plural_ | " +
          generate(omorfi, word, upos,
                   "[VOICE=ACT][MOOD=INDV][TENSE=PRESENT][NEG=CON]") +
          " | " +
          generate(omorfi, word, upos,
                   "[VOICE=ACT][MOOD=INDV][TENSE=PAST][NUM=PL][NEG=CON]") +
          " | " +
          generate(omorfi, word, upos,
                   "[VOICE=ACT][MOOD=COND][NEG=CON]") +
          " | " +
          generate(omorfi, word, upos,
                   "[VOICE=ACT][MOOD=POTN][NEG=CON]") +
          " | " +
          generate(omorfi, word, upos,
                   "[VOICE=ACT][MOOD=IMPV][NEG=CON]") + " |",
          file=outfile)
    print("| _Connegative impersonal_ | " +
          generate(omorfi, word, upos,
                   "[VOICE=PSS][MOOD=INDV][TENSE=PRESENT][PERS=PE4]" +
                   "[NEG=CON]") +
          " | " +
          generate(omorfi, word, upos,
                   "[VOICE=PSS][MOOD=INDV][TENSE=PAST][PERS=PE4]" +
                   "[NEG=CON]") +
          " | " +
          generate(omorfi, word, upos,
                   "[VOICE=PSS][MOOD=COND][PERS=PE4][NEG=CON]") +
          " | " +
          generate(omorfi, word, upos,
                   "[VOICE=PSS][MOOD=POTN][PERS=PE4][NEG=CON]") +
          " | " +
          generate(omorfi, word, upos,
                   "[VOICE=PSS][MOOD=IMPV][PERS=PE4][NEG=CON]") + " |",
          file=outfile)


def print_infinites(omorfi, word, upos, outfile):
    print("#### Non-finite forms", file=outfile)
    print(file=outfile)
    print("| *Form* | _Active_ | _Passive_ | _Possessive_ | _Clitic_ |",
          file=outfile)
    print("|:-------|:---------|:----------|:-------------|:---------|",
          file=outfile)
    print("| _infinitive A lative_ | " +
          generate(omorfi, word, upos,
                   "[VOICE=ACT][INF=A][NUM=SG][CASE=LAT]") + " | " +
          generate(omorfi, word, upos,
                   "[VOICE=PSS][INF=A][NUM=SG][CASE=LAT]") + " | " +
          generate(omorfi, word, upos,
                   "[VOICE=ACT][INF=A][NUM=SG][CASE=LAT][POSS=SG1]") +
          " | " +
          generate(omorfi, word, upos,
                   "[VOICE=ACT][INF=A][NUM=SG][CASE=LAT][CLIT=KO]") + " |",
          file=outfile)
    print("| _infinitive A translative_ | " +
          generate(omorfi, word, upos,
                   "[VOICE=ACT][INF=A][CASE=TRA]") + " | " +
          generate(omorfi, word, upos,
                   "[VOICE=PSS][INF=A][CASE=TRA]") + " | " +
          generate(omorfi, word, upos,
                   "[VOICE=ACT][INF=A][CASE=TRA][POSS=SG1]") +
          " | " +
          generate(omorfi, word, upos,
                   "[VOICE=ACT][INF=A][CASE=TRA][CLIT=KO]") + " |",
          file=outfile)
    print("| _infinitive E inessive_ | " +
          generate(omorfi, word, upos,
                   "[VOICE=ACT][INF=E][NUM=SG][CASE=INE]") + " | " +
          generate(omorfi, word, upos,
                   "[VOICE=PSS][INF=E][NUM=SG][CASE=INE]") + " | " +
          generate(omorfi, word, upos,
                   "[VOICE=ACT][INF=E][NUM=SG][CASE=INE][POSS=SG1]") +
          " | " +
          generate(omorfi, word, upos,
                   "[VOICE=PSS][INF=E][NUM=SG][CASE=INE][CLIT=KO]") + " |",
          file=outfile)
    print("| _infinitive E instructive_ | " +
          generate(omorfi, word, upos,
                   "[VOICE=ACT][INF=E][NUM=SG][CASE=INS]") + " | " +
          generate(omorfi, word, upos,
                   "[VOICE=PSS][INF=E][NUM=SG][CASE=INS]") + " | " +
          generate(omorfi, word, upos,
                   "[VOICE=ACT][INF=E][NUM=SG][CASE=INS][POSS=SG1]") +
          " | " +
          generate(omorfi, word, upos,
                   "[VOICE=PSS][INF=E][NUM=SG][CASE=INS][CLIT=KO]") + " |",
          file=outfile)
    print("| _infinitive MA inessive_ | " +
          generate(omorfi, word, upos,
                   "[VOICE=ACT][INF=MA][NUM=SG][CASE=INE]") + " | " +
          generate(omorfi, word, upos,
                   "[VOICE=PSS][INF=MA][NUM=SG][CASE=INE]") + " | " +
          generate(omorfi, word, upos,
                   "[VOICE=ACT][INF=MA][NUM=SG][CASE=INE][POSS=SG1]") +
          " | " +
          generate(omorfi, word, upos,
                   "[VOICE=PSS][INF=MA][NUM=SG][CASE=INE][CLIT=KO]") + " |",
          file=outfile)
    print("| _infinitive MA elative_ | " +
          generate(omorfi, word, upos,
                   "[VOICE=ACT][INF=MA][NUM=SG][CASE=ELA]") + " | " +
          generate(omorfi, word, upos,
                   "[VOICE=PSS][INF=MA][NUM=SG][CASE=ELA]") + " | " +
          generate(omorfi, word, upos,
                   "[VOICE=ACT][INF=MA][NUM=SG][CASE=ELA][POSS=SG1]") +
          " | " +
          generate(omorfi, word, upos,
                   "[VOICE=PSS][INF=MA][NUM=SG][CASE=ELA][CLIT=KO]") + " |",
          file=outfile)
    print("| _infinitive MA illative_ | " +
          generate(omorfi, word, upos,
                   "[VOICE=ACT][INF=MA][NUM=SG][CASE=ILL]") + " | " +
          generate(omorfi, word, upos,
                   "[VOICE=PSS][INF=MA][NUM=SG][CASE=ILL]") + " | " +
          generate(omorfi, word, upos,
                   "[VOICE=ACT][INF=MA][NUM=SG][CASE=ILL][POSS=SG1]") +
          " | " +
          generate(omorfi, word, upos,
                   "[VOICE=PSS][INF=MA][NUM=SG][CASE=ILL][CLIT=KO]") + " |",
          file=outfile)
    print("| _infinitive MA adessive_ | " +
          generate(omorfi, word, upos,
                   "[VOICE=ACT][INF=MA][NUM=SG][CASE=ADE]") + " | " +
          generate(omorfi, word, upos,
                   "[VOICE=PSS][INF=MA][NUM=SG][CASE=ADE]") + " | " +
          generate(omorfi, word, upos,
                   "[VOICE=ACT][INF=MA][NUM=SG][CASE=ADE][POSS=SG1]") +
          " | " +
          generate(omorfi, word, upos,
                   "[VOICE=PSS][INF=MA][NUM=SG][CASE=ADE][CLIT=KO]") + " |",
          file=outfile)
    print("| _infinitive MA abessive_ | " +
          generate(omorfi, word, upos,
                   "[VOICE=ACT][INF=MA][NUM=SG][CASE=ABE]") + " | " +
          generate(omorfi, word, upos,
                   "[VOICE=PSS][INF=MA][NUM=SG][CASE=ABE]") + " | " +
          generate(omorfi, word, upos,
                   "[VOICE=ACT][INF=MA][NUM=SG][CASE=ABE][POSS=SG1]") +
          " | " +
          generate(omorfi, word, upos,
                   "[VOICE=PSS][INF=MA][NUM=SG][CASE=ABE][CLIT=KO]") + " |",
          file=outfile)
    print("| _infinitive MA instructive_ | " +
          generate(omorfi, word, upos,
                   "[VOICE=ACT][INF=MA][NUM=SG][CASE=INS]") + " | " +
          generate(omorfi, word, upos,
                   "[VOICE=PSS][INF=MA][NUM=SG][CASE=INS]") + " | " +
          generate(omorfi, word, upos,
                   "[VOICE=ACT][INF=MA][NUM=SG][CASE=INS][POSS=SG1]") +
          " | " +
          generate(omorfi, word, upos,
                   "[VOICE=PSS][INF=MA][NUM=SG][CASE=INS][CLIT=KO]") + " |",
          file=outfile)
    print("| _derivation MINEN_ | " +
          generate(omorfi, word, upos,
                   "[INF=MINEN]") + " | " +
          generate(omorfi, word, upos,
                   "[VOICE=PSS][INF=MINEN]") + " | " +
          generate(omorfi, word, upos,
                   "[DRV=MINEN][NUM=SG][CASE=NOM][POSS=SG1]") +
          " | " +
          generate(omorfi, word, upos,
                   "[DRV=MINEN][NUM=SG][CASE=NOM][CLIT=KO]") + " |",
          file=outfile)
    print("| _derivation MAISILLA_ | " +
          generate(omorfi, word, upos,
                   "[VOICE=ACT][DRV=MAISILLA]") + " | " +
          generate(omorfi, word, upos,
                   "[VOICE=PSS][DRV=MAISILLA]") + " | " +
          generate(omorfi, word, upos,
                   "[VOICE=ACT][DRV=MAISILLA][POSS=SG1]") +
          " | " +
          generate(omorfi, word, upos,
                   "[VOICE=PSS][DRV=MAISILLA][CLIT=KO]") + " |",
          file=outfile)
    print("| _participle NUT_ | " +
          generate(omorfi, word, upos,
                   "[VOICE=ACT][PCP=NUT][CMP=POS][NUM=SG][CASE=NOM]") +
          " | " +
          generate(omorfi, word, upos,
                   "[VOICE=PSS][PCP=NUT][CMP=POS][NUM=SG][CASE=NOM]") +
          " | " +
          generate(omorfi, word, upos,
                   "[VOICE=ACT][PCP=NUT][CMP=POS][NUM=SG][CASE=INE]" +
                   "[POSS=SG1]") + " | " +
          generate(omorfi, word, upos,
                   "[VOICE=PSS][PCP=NUT][CMP=POS][NUM=SG][CASE=ADE][" +
                   "CLIT=KO]") + " |", file=outfile)
    print("| _participle VA_ | " +
          generate(omorfi, word, upos,
                   "[VOICE=ACT][PCP=VA][CMP=POS][NUM=SG][CASE=NOM]") +
          " | " +
          generate(omorfi, word, upos,
                   "[VOICE=PSS][PCP=VA][CMP=POS][NUM=SG][CASE=NOM]") +
          " | " +
          generate(omorfi, word, upos,
                   "[VOICE=ACT][PCP=VA][CMP=POS][NUM=SG][CASE=ELA]" +
                   "[POSS=SG1]") + " | " +
          generate(omorfi, word, upos,
                   "[VOICE=PSS][PCP=VA][CMP=POS][NUM=SG][CASE=ABL][" +
                   "CLIT=KO]") + " |", file=outfile)
    print("| _participle MA_ | " +
          generate(omorfi, word, upos,
                   "[PCP=AGENT][CMP=POS][NUM=SG][CASE=NOM]") +
          " | " +
          generate(omorfi, word, upos,
                   "[VOICE=PSS][PCP=AGENT][CMP=POS][NUM=SG][CASE=NOM]") +
          " | " +
          generate(omorfi, word, upos,
                   "[PCP=AGENT][CMP=POS][NUM=SG][CASE=ELA]" +
                   "[POSS=SG1]") + " | " +
          generate(omorfi, word, upos,
                   "[PCP=AGENT][CMP=POS][NUM=SG][CASE=ABL][" +
                   "CLIT=KO]") + " |", file=outfile)
    print("| _participle MATON_ | " +
          generate(omorfi, word, upos,
                   "[PCP=NEG][CMP=POS][NUM=SG][CASE=NOM]") +
          " | " +
          generate(omorfi, word, upos,
                   "[VOICE=PSS][PCP=NEG][CMP=POS][NUM=SG][CASE=NOM]") +
          " | " +
          generate(omorfi, word, upos,
                   "[PCP=NEG][CMP=POS][NUM=SG][CASE=ELA]" +
                   "[POSS=SG1]") + " | " +
          generate(omorfi, word, upos,
                   "[PCP=NEG][CMP=POS][NUM=SG][CASE=ABL][" +
                   "CLIT=KO]") + " |", file=outfile)


def main():
    """Command-line interface for omorfi's sort | uniq -c tester."""
    a = ArgumentParser()
    a.add_argument('-g', '--generator', metavar='FSAFILE', required=True,
                   help="load generator from FSAFILE")
    a.add_argument('-w', '--word', metavar="WORD_ID", required=True,
                   help="generate forms of word WORD_ID")
    a.add_argument('-o', '--output', metavar="OUTFILE",
                   type=FileType('w'),
                   dest="outfile", help="log outputs to OUTFILE")
    a.add_argument('-X', '--statistics', metavar="STATFILE",
                   type=FileType('w'),
                   dest="statfile", help="statistics")
    a.add_argument('-v', '--verbose', action="store_true", default=False,
                   help="Print verbosely while processing")
    a.add_argument('-O', '--output-format', metavar="OFORMAT",
                   default="markdown",
                   help="Create output table in OFORMAT")
    a.add_argument('-u', '--upos', metavar="UPOS", required=True,
                   choices=["ADJ", "NOUN", "VERB", "NUM", "X"],
                   help="generate inflection table for UPOS")
    options = a.parse_args()
    omorfi = Omorfi(options.verbose)
    try:
        if options.generator:
            if options.verbose:
                print("reading generator from", options.generator)
            omorfi.load_generator(options.generator)
        if not options.statfile:
            options.statfile = stdout
        if not options.outfile:
            options.outfile = stdout
    except IOError:
        print("Could not process file", options.generator, file=stderr)
        exit(2)
    # for make check target
    realstart = perf_counter()
    cpustart = process_time()
    print("### Inflection of", options.word, file=options.outfile)
    print(file=options.outfile)
    if options.upos == 'NOUN':
        print_nominals(omorfi, options.word, options.upos, options.outfile)
    elif options.upos == 'ADJ':
        print_comparatives(omorfi, options.word, options.upos, 'POS',
                           options.outfile)
        # comparisons
        print(file=options.outfile)
        print_comparatives(omorfi, options.word, options.upos, "CMP",
                           options.outfile)
        print(file=options.outfile)
        print_comparatives(omorfi, options.word, options.upos, "SUP",
                           options.outfile)
    elif options.upos == 'NUM':
        print_numerals(omorfi, options.word, options.upos, options.outfile)
    elif options.upos == 'VERB':
        print_finites(omorfi, options.word, options.upos, options.outfile)
        print(file=options.outfile)
        print_infinites(omorfi, options.word, options.upos, options.outfile)
    print(file=options.outfile)
    print("_Note:_ the inflection tables cover small percentage of the " +
          "whole inflectional paradigm, for full list, see [" +
          options.word + " full form list](" + options.word + ".html)",
          file=options.outfile)
    print(file=options.outfile)
    realend = perf_counter()
    cpuend = process_time()
    print("CPU time:", cpuend - cpustart, "real time:", realend - realstart)
    exit(0)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Functions to format apertium style analyses from omorfi data."""

# Author: Omorfi contributors <omorfi-devel@groups.google.com> 2015

#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#
# utils to format apertium style data from omorfi database values

from .formatter import Formatter
from ..settings import optional_hyphen, word_boundary
from ..string_manglers import lexc_escape


class NoTagsFormatter(Formatter):
    """A formatter that can create tagless analysers. This is used for e.g.,
    lemmatisers, acceptors, and morph segmenters."""

    def __init__(self, verbose=True, **kwargs):
        """Create omorfi tag converter for tagless automata with given
        verbosity and options.

        @param verbose set to false to disable printing
        @param segment set to true to create segmenter
        @param lemmatise set to true to create lemmatiser
        """
        ## verbosity
        self.verbose = verbose
        ## if we are creating a segmenter
        self.segment = False
        ## if we are creating a lemmatiser
        self.lemmatise = False
        if 'lemmatise' in kwargs and kwargs['lemmatise']:
            self.lemmatise = True
        elif 'segment' in kwargs and kwargs['segment']:
            self.segment = True

    def stuff2lexc(self, stuff):
        """Convert omorfi analyses to nothing

        @return empty string unless relevant for lemmatiser or segmenter
        """
        if stuff == 'Bc':
            return word_boundary
        else:
            return ""

    def analyses2lexc(self, anals):
        """Convert omorfi analyses to nothing

        @return empty string unless relevant for lemmatiser or segmenter
        """
        apestring = ''
        for i in anals.split('|'):
            apestring += self.stuff2lexc(i)
        return apestring

    def continuation2lexc(self, anals, surf, cont):
        """Create lexc content for omorfi continuation class data

        @return lexc-formatted continuation class entry
        """
        analstring = self.analyses2lexc(anals)
        # the followings have surface fragments in continuations
        if 'DIGITS_' in cont and not ('BACK' in cont or 'FRONT' in cont):
            analstring = lexc_escape(surf) + analstring
        elif 'PUNCT_NONSTD_EXCL_LOOP' in cont:
            analstring = lexc_escape(surf) + analstring
        elif self.segment:
            analstring = lexc_escape(surf)
        elif self.lemmatise:
            pass
        else:
            analstring = lexc_escape(surf)
        return "%s:%s\t%s ;\n" % (analstring, lexc_escape(surf), cont)

    def wordmap2lexc(self, wordmap):
        """Format omorfi lexical item for tagless analysis

        @return lexc-format entry for lemma/stub analysis
        """
        if wordmap['lemma'] == ' ':
            return ''
        if self.lemmatise:
            wordmap['analysis'] = lexc_escape(wordmap['lemma'])
        elif self.segment:
            wordmap['analysis'] = lexc_escape(wordmap['stub'])
        else:
            wordmap['analysis'] = lexc_escape(wordmap['stub'])
        wordmap['stub'] = wordmap['stub'].replace(
            word_boundary, optional_hyphen)
        wordmap['stub'] = lexc_escape(wordmap['stub'])
        lexc_line = "%s:%s\t%s\t;\n" % (wordmap['analysis'], wordmap['stub'],
                                        wordmap['new_para'])
        if 'BLACKLISTED' in wordmap['new_para']:
            return "! ! !" + lexc_line
        else:
            return lexc_line

    def multichars_lexc(self):
        """Create analysis tags for lemmas and segments

        @return lexc-formatted string for Multichar_symbols block without
                analyses
        """
        multichars = "Multichar_Symbols\n"
        multichars += Formatter.multichars_lexc(self)
        return multichars

    def root_lexicon_lexc(self):
        """Create lexc root lexicon for lemmatisers and segmenters

        @return lexc-format string for Root lexicon of lemmatiser
        """
        root = Formatter.root_lexicon_lexc(self)
        return root


# self test
if __name__ == '__main__':
    formatter = NoTagsFormatter()
    exit(0)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Functions to format LMF compatible XML from omorfi data."""

# Author: Omorfi contributors <omorfi-devel@groups.google.com> 2015

#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.

from xml.sax.saxutils import escape as xml_escape

from ftb3_formatter import Ftb3Formatter

from .settings import version_id_easter_egg


def make_xmlid(s):
    return s.replace("?", "_UNK").replace("→", "_right").replace("←", "left").replace(".", "_")


def format_multichars_lexc_xml():
    multichars = "  <Multichar_Symbols>\n"
    for key, value in Ftb3Formatter.stuff2ftb3.items():
        key = make_xmlid(key)
        if key != '':
            if value != '':
                multichars += "    <mcs id='" + key + \
                    "'>" + xml_escape(value) + "</mcs>\n"
            else:
                multichars += "    <mcs id='" + key + "'>" + key + "</mcs>\n"
        else:
            pass

    multichars += """<!-- Following specials exist in all versions of omorfi -->
    <mcs id="hyph">{hyph?}</mcs>
    <mcs id="deriv">»</mcs>
    <mcs id="infl">&gt;</mcs>
    <mcs id="wb">|</mcs>
    """
    multichars += "    <mcs id='VERSION'>" + version_id_easter_egg + '</mcs>\n'
    multichars += "  </Multichar_Symbols>"
    return multichars


def format_root_lexicon_xml():
    root = '  <LEXICON id="Root">\n'
    root += """<!-- ... -->
    <e><a/><i/><cont lexica="NOUNS ADJECTIVES VERBS NUMERALS DIGITSS ACRONYMS PRONOUNS PARTICLES PUNCTUATIONS FIFTYONE"/></e>
"""
    root += '    <e><a><s mcs="B_right"/></a><i>-</i><cont lexica="NOUN ADJECTIVE SUFFIX"/></e>\n'
    root += '    <e><a><s mcs="VERSION"/></a><i/><cont lexica="_end"/></e>\n'
    root += '  </LEXICON>\n'
    return root


def format_lexc_xml(wordmap):
    analysis = xml_escape(wordmap['lemma'])
    analysis = analysis.replace(
        '|', '<s mcs="wb"/>').replace('_', '<s mcs="mb"/>')
    analysis += '<s mcs="' + wordmap['pos'] + '"/>'
    if wordmap['is_proper']:
        analysis += '<s mcs="proper"/>'
    if wordmap['is_suffix']:
        analysis = "<s mcs='suffix'/>" + analysis
    if wordmap['is_prefix']:
        analysis += "<s mcs='prefix'/>"
    stub = xml_escape(wordmap['stub'])
    stub = stub.replace('|', '<s mcs="wb"/>').replace('_', '<s mcs="mb"/>')
    return ('    <e><a>%s</a><i>%s</i><cont lexica="%s"/></e>' %
            (analysis, stub, wordmap['new_para']))


def format_continuation_lexicon_xml(tsvparts):
    xmlstring = '    <e>'
    if tsvparts[1] != '':
        xmlstring += '<a>'
        for anal in tsvparts[1].split('|'):
            if anal in Ftb3Formatter.stuff2ftb3:
                anal = make_xmlid(anal)
                xmlstring += '<s mcs="' + anal + '"/>'
            else:
                xmlstring += xml_escape(anal)
        xmlstring += '</a>'
    else:
        xmlstring += '<a/>'
    xmlstring += "<i>" + xml_escape(tsvparts[2]) + "</i>"
    xmlstring += '<cont lexica="' + \
        " ".join(tsvparts[3:]).replace("#", "_END") + '"/></e>\n'
    return xmlstring
